"""
Copyright 2016 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import asyncio
import datetime
import logging
import typing
import uuid
from asyncio import subprocess
from typing import Optional
from unittest.mock import Mock
from uuid import UUID, uuid4

import pytest
from tornado.httpclient import AsyncHTTPClient

from inmanta import config, const, data
from inmanta.config import Config
from inmanta.const import AgentAction, AgentStatus
from inmanta.protocol import Result, handle, typedmethod
from inmanta.protocol.common import ReturnValue
from inmanta.server import SLICE_AGENT_MANAGER, SLICE_AUTOSTARTED_AGENT_MANAGER, protocol
from inmanta.server.agentmanager import AgentManager, AutostartedAgentManager, SessionAction, SessionManager
from inmanta.server.bootloader import InmantaBootloader
from inmanta.server.protocol import ServerSlice, Session
from utils import UNKWN, NullAgent, assert_equal_ish, retry_limited

LOGGER = logging.getLogger(__name__)


@pytest.fixture(scope="function")
async def auto_start_agent():
    # In this file, we allow auto started agents
    return True


class Collector:
    def __init__(self):
        self.values = []

    def __call__(self, arg):
        self.values.append(arg)

    async def proccess(self):
        while len(self.values) > 0:
            x = self.values
            self.values = []
            await asyncio.gather(*x)


async def empty_future(*args, **kwargs):
    pass


async def api_call_future(*args, **kwargs) -> Result:
    """
    Mock implementation of Client methods
    """
    client = Mock()
    method_properties = Mock()
    return Result(200, "X", client=client, method_properties=method_properties)


class MockSession:
    """
    An environment that segments agents connected to the server
    """

    def __init__(self, sid, tid, endpoint_names: set[str], nodename):
        self._sid = sid
        self.tid = tid
        self.endpoint_names = endpoint_names
        self.nodename = nodename
        self.client = Mock()
        self.client.set_state.side_effect = empty_future
        self.client.get_status = api_call_future

    def get_id(self):
        return self._sid

    id = property(get_id)

    def get_client(self):
        return self.client

    async def expire_and_abort(self, timeout: float):
        # Only called on teardown
        pass


async def assert_agent_state(env_id: UUID, name: str, state: AgentStatus, sid: Optional[UUID]) -> None:
    agent = await data.Agent.get(env_id, name)

    assert agent.get_status() == state
    if state == AgentStatus.paused:
        assert agent.primary is None
        assert agent.paused
    elif state == AgentStatus.down:
        assert agent.primary is None
        assert not agent.paused
    elif state == AgentStatus.up:
        assert agent.primary is not None
        agent_instance = await data.AgentInstance.get_by_id(agent.primary)
        agent_proc = await data.AgentProcess.get_one(sid=agent_instance.process)
        assert agent_proc.sid == sid
        assert agent.get_status() == AgentStatus.up


async def assert_state_agents(
    env_id: UUID,
    s1: AgentStatus,
    s2: AgentStatus,
    s3: AgentStatus,
    sid1: Optional[UUID] = None,
    sid2: Optional[UUID] = None,
    sid3: Optional[UUID] = None,
) -> None:
    await assert_agent_state(env_id, "agent1", s1, sid1)
    await assert_agent_state(env_id, "agent2", s2, sid2)
    await assert_agent_state(env_id, "agent3", s3, sid3)


def assert_state_agents_retry(
    env_id: UUID,
    s1: AgentStatus,
    s2: AgentStatus,
    s3: AgentStatus,
    sid1: Optional[UUID] = None,
    sid2: Optional[UUID] = None,
    sid3: Optional[UUID] = None,
) -> typing.Callable[[], None]:
    async def func() -> bool:
        try:
            await assert_state_agents(env_id, s1, s2, s3, sid1, sid2, sid3)
        except AssertionError:
            return False
        return True

    return func


@pytest.mark.parametrize("auto_start_agent", [False])  # prevent autostart to keep agent under control
async def test_primary_selection(server, environment):
    env_id = UUID(environment)
    env = await data.Environment.get_by_id(env_id)
    am = server.get_slice(SLICE_AGENT_MANAGER)

    # second env to detect crosstalk
    project = data.Project(name="test_x")
    await project.insert()
    env2 = data.Environment(name="testenv_x", project=project.id)
    await env2.insert()

    await data.Agent(environment=env.id, name="agent1", paused=True).insert()
    await data.Agent(environment=env.id, name="agent2", paused=False).insert()
    await data.Agent(environment=env.id, name="agent3", paused=False).insert()

    # one session
    ts1 = MockSession(uuid4(), env.id, {"agent1", "agent2"}, "ts1")
    await am.new_session(ts1, set(ts1.endpoint_names))
    await am._session_listener_actions.join()
    assert len(am.sessions) == 1

    # cross talk session
    ts2 = MockSession(uuid4(), env2.id, {"agent1", "agent2"}, "ts2")
    await am.new_session(ts2, set(ts2.endpoint_names))
    await am._session_listener_actions.join()
    assert len(am.sessions) == 2

    ts1.get_client().set_state.assert_called_with("agent2", enabled=True)
    ts1.get_client().reset_mock()
    await retry_limited(
        assert_state_agents_retry(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.down, sid2=ts1.id), 10
    )

    # test is_primary
    assert am.is_primary(env, ts1.id, "agent2")
    assert not am.is_primary(env, ts1.id, "agent1")
    assert not am.is_primary(env, uuid4(), "agent2")

    # alive
    await am.seen(ts1, set(ts1.endpoint_names))
    await am._session_listener_actions.join()
    assert len(am.sessions) == 2
    await retry_limited(
        assert_state_agents_retry(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.down, sid2=ts1.id), 10
    )

    # second session
    ts2 = MockSession(uuid4(), env.id, {"agent3", "agent2"}, "ts2")
    await am.new_session(ts2, set(ts2.endpoint_names))
    await am._session_listener_actions.join()
    assert len(am.sessions) == 3
    ts2.get_client().set_state.assert_called_with("agent3", enabled=True)
    ts2.get_client().reset_mock()
    await retry_limited(
        assert_state_agents_retry(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.up, sid2=ts1.id, sid3=ts2.id), 10
    )

    # test is_primary
    assert not am.is_primary(env, ts2.id, "agent1")
    assert not am.is_primary(env, ts1.id, "agent1")
    assert am.is_primary(env, ts1.id, "agent2")
    assert am.is_primary(env, ts2.id, "agent3")

    # expire first
    await am.expire(ts1, set(ts1.endpoint_names))
    await am._session_listener_actions.join()
    assert len(am.sessions) == 2
    ts2.get_client().set_state.assert_called_with("agent2", enabled=True)
    ts2.get_client().reset_mock()
    await retry_limited(
        assert_state_agents_retry(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.up, sid2=ts2.id, sid3=ts2.id), 10
    )

    # expire second
    await am.expire(ts2, set(ts2.endpoint_names))
    await am._session_listener_actions.join()
    assert len(am.sessions) == 1
    await retry_limited(assert_state_agents_retry(env.id, AgentStatus.paused, AgentStatus.down, AgentStatus.down), 10)

    # test is_primary
    assert not am.is_primary(env, ts1.id, "agent2")
    assert not am.is_primary(env, ts2.id, "agent3")


async def test_api(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="testenv", project=project.id)
    await env.insert()
    env2 = data.Environment(name="testenv2", project=project.id)
    await env2.insert()
    env3 = data.Environment(name="testenv3", project=project.id)
    await env3.insert()
    # Exact replica of one to detect crosstalk
    env4 = data.Environment(name="testenv4", project=project.id)
    await env4.insert()
    env5 = data.Environment(name="testenv5", project=project.id)
    await env5.insert()

    await data.Agent(environment=env.id, name="agent1", paused=True).insert()
    await data.Agent(environment=env.id, name="agent2", paused=False).insert()
    await data.Agent(environment=env.id, name="agent3", paused=False).insert()
    await data.Agent(environment=env2.id, name="agent4", paused=False).insert()
    await data.Agent(environment=env4.id, name="agent1", paused=True).insert()
    await data.Agent(environment=env4.id, name="agent2", paused=False).insert()
    await data.Agent(environment=env4.id, name="agent3", paused=False).insert()
    await data.Agent(environment=env5.id, name="agent1", paused=False).insert()

    server = Mock()
    futures = Collector()
    server.add_background_task.side_effect = futures
    am = AgentManager(server, False)
    am.add_background_task = futures
    am.running = True

    # one session
    ts1 = MockSession(uuid4(), env.id, {"agent1", "agent2"}, "ts1")
    await am._register_session(ts1, set(ts1.endpoint_names), datetime.datetime.now())
    # second session
    ts2 = MockSession(uuid4(), env.id, ["agent3", "agent2"], "ts2")
    await am._register_session(ts2, set(ts2.endpoint_names), datetime.datetime.now())
    # third session
    ts3 = MockSession(uuid4(), env3.id, ["agentx"], "ts3")
    await am._register_session(ts3, set(ts3.endpoint_names), datetime.datetime.now())
    # fourth session
    ts4 = MockSession(uuid4(), env4.id, {"agent1", "agent2"}, "ts4")
    await am._register_session(ts4, set(ts4.endpoint_names), datetime.datetime.now())
    # fifth session
    ts5 = MockSession(uuid4(), env5.id, {"agent1"}, "ts5")
    await am._register_session(ts5, set(ts5.endpoint_names), datetime.datetime.now())

    await futures.proccess()
    assert len(am.sessions) == 5

    # Getting all non expired agent processes
    code, all_agents_processes = await am.list_agent_processes(environment=None, expired=False)
    assert code == 200
    assert len(all_agents_processes["processes"]) == 5

    # Getting all agent processes (including expired)
    code, all_agents_processes = await am.list_agent_processes(environment=None, expired=True)
    assert code == 200
    assert len(all_agents_processes["processes"]) == 5

    # Making fifth session expire
    expiration = datetime.datetime.now().astimezone()
    await am._expire_session(ts5, set(ts5.endpoint_names), expiration)

    await futures.proccess()
    assert len(am.sessions) == 4

    # Getting all non expired agent processes
    code, all_agents_processes = await am.list_agent_processes(environment=None, expired=False)
    assert code == 200
    assert len(all_agents_processes["processes"]) == 4
    for agent_process in all_agents_processes["processes"]:
        assert agent_process["expired"] is None

    # Getting all agent processes (including expired)
    code, all_agents_processes = await am.list_agent_processes(environment=None, expired=True)
    assert code == 200
    assert len(all_agents_processes["processes"]) == 5

    shouldbe = {
        "processes": [
            {
                "first_seen": UNKWN,
                "expired": None,
                "hostname": "ts1",
                "last_seen": UNKWN,
                "endpoints": [
                    {"id": UNKWN, "name": "agent1", "process": UNKWN},
                    {"id": UNKWN, "name": "agent2", "process": UNKWN},
                ],
                "environment": env.id,
            },
            {
                "first_seen": UNKWN,
                "expired": None,
                "hostname": "ts2",
                "last_seen": UNKWN,
                "endpoints": [
                    {"id": UNKWN, "name": "agent2", "process": UNKWN},
                    {"id": UNKWN, "name": "agent3", "process": UNKWN},
                ],
                "environment": env.id,
            },
            {
                "first_seen": UNKWN,
                "expired": None,
                "hostname": "ts3",
                "last_seen": UNKWN,
                "endpoints": [{"id": UNKWN, "name": "agentx", "process": UNKWN}],
                "environment": env3.id,
            },
            {
                "first_seen": UNKWN,
                "expired": None,
                "hostname": "ts4",
                "last_seen": UNKWN,
                "endpoints": [
                    {"id": UNKWN, "name": "agent1", "process": UNKWN},
                    {"id": UNKWN, "name": "agent2", "process": UNKWN},
                ],
                "environment": env4.id,
            },
            {
                "first_seen": UNKWN,
                "expired": expiration,
                "hostname": "ts5",
                "last_seen": UNKWN,
                "endpoints": [
                    {"expired": expiration, "id": UNKWN, "name": "agent1", "process": UNKWN},
                ],
                "environment": env5.id,
            },
        ]
    }

    assert_equal_ish(shouldbe, all_agents_processes, sortby=["hostname", "name"])
    # There is 5 agent processes, we take the 3rd one and will select the two before, and two after it.
    agentid = sorted(all_agents_processes["processes"], key=lambda p: p["sid"])[2]["sid"]

    start = agentid
    code, all_agents_processes = await am.list_agent_processes(environment=None, expired=True, start=start)
    assert code == 200
    assert len(all_agents_processes["processes"]) == 2
    for agent_process in all_agents_processes["processes"]:
        assert (
            agent_process["sid"] > start
        ), f"List of agent processes should not contain a sid (={agent_process['sid']}) before or equal to start (={start})"

    end = agentid
    code, all_agents_processes = await am.list_agent_processes(environment=None, expired=True, end=end)
    assert code == 200
    assert len(all_agents_processes["processes"]) == 2
    for agent_process in all_agents_processes["processes"]:
        assert (
            agent_process["sid"] < end
        ), f"List of agent processes should not contain a sid (={agent_process['sid']}) after or equal to end (={end})"

    code, all_agents = await am.list_agent_processes(environment=env.id, expired=False)
    assert code == 200
    agentid1 = all_agents["processes"][0]["sid"]
    agentid2 = all_agents["processes"][1]["sid"]

    shouldbe = {
        "processes": [
            {
                "first_seen": UNKWN,
                "expired": None,
                "hostname": "ts1",
                "last_seen": UNKWN,
                "endpoints": [
                    {"id": UNKWN, "name": "agent1", "process": UNKWN},
                    {"id": UNKWN, "name": "agent2", "process": UNKWN},
                ],
                "environment": env.id,
            },
            {
                "first_seen": UNKWN,
                "expired": None,
                "hostname": "ts2",
                "last_seen": UNKWN,
                "endpoints": [
                    {"id": UNKWN, "name": "agent2", "process": UNKWN},
                    {"id": UNKWN, "name": "agent3", "process": UNKWN},
                ],
                "environment": env.id,
            },
        ]
    }

    assert_equal_ish(shouldbe, all_agents, sortby=["hostname", "name"])

    code, all_agents = await am.list_agent_processes(environment=env2.id, expired=False)
    assert code == 200

    shouldbe = {"processes": []}

    assert_equal_ish(shouldbe, all_agents)

    report = await am.get_agent_process_report(agentid1)
    assert (200, "X") == report

    report = await am.get_agent_process_report(agentid2)
    assert (200, "X") == report

    result, _ = await am.get_agent_process_report(uuid4())
    assert result == 404

    code, all_agents = await am.list_agents(None)
    assert code == 404

    start = "agent2"
    code, all_agents = await am.list_agents(env=env, start=start)
    assert code == 200
    assert len(all_agents["agents"]) == 1
    for a in all_agents["agents"]:
        assert a["name"] > start, f"List of agent should not contain a name (={a['name']}) before or equal to start (={start})"

    end = "agent2"
    code, all_agents = await am.list_agents(env=env, end=end)
    assert code == 200
    assert len(all_agents["agents"]) == 1
    for a in all_agents["agents"]:
        assert a["name"] < end, f"List of agent should not contain a name (={a['name']}) after or equal to end (={end})"

    code, all_agents = await am.list_agents(env2)
    assert code == 200
    shouldbe = {
        "agents": [
            {"name": "agent4", "paused": False, "last_failover": "", "primary": "", "environment": env2.id, "state": "down"}
        ]
    }
    assert_equal_ish(shouldbe, all_agents)


async def test_expire_all_sessions_in_db(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="testenv", project=project.id)
    await env.insert()
    await data.Agent(environment=env.id, name="agent1", paused=True).insert()
    await data.Agent(environment=env.id, name="agent2", paused=False).insert()
    await data.Agent(environment=env.id, name="agent3", paused=False).insert()

    server = Mock()
    futures = Collector()
    server.add_background_task.side_effect = futures
    am = AgentManager(server, False)
    am.add_background_task = futures
    am.running = True

    # one session
    ts1 = MockSession(uuid4(), env.id, {"agent1", "agent2"}, "ts1")
    await am._register_session(ts1, set(ts1.endpoint_names), datetime.datetime.now())
    await futures.proccess()
    assert len(am.sessions) == 1
    ts1.get_client().set_state.assert_called_with("agent2", enabled=True)
    ts1.get_client().reset_mock()
    await assert_state_agents(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.down, sid2=ts1.id)

    # alive
    await am._seen_session(ts1, set(ts1.endpoint_names))
    await futures.proccess()
    assert len(am.sessions) == 1
    await assert_state_agents(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.down, sid2=ts1.id)

    # second session
    ts2 = MockSession(uuid4(), env.id, {"agent3", "agent2"}, "ts2")
    await am._register_session(ts2, set(ts2.endpoint_names), datetime.datetime.now())
    await futures.proccess()
    assert len(am.sessions) == 2
    ts2.get_client().set_state.assert_called_with("agent3", enabled=True)
    ts2.get_client().reset_mock()
    await assert_state_agents(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.up, sid2=ts1.id, sid3=ts2.id)

    # expire first
    await am._expire_session(ts1, set(ts1.endpoint_names), datetime.datetime.now())
    await futures.proccess()
    assert len(am.sessions) == 1
    ts2.get_client().set_state.assert_called_with("agent2", enabled=True)
    ts2.get_client().reset_mock()
    await assert_state_agents(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.up, sid2=ts2.id, sid3=ts2.id)

    # failover
    am = AgentManager(server, False)
    am.add_background_task = futures
    am.running = True
    await am._expire_all_sessions_in_db()
    await assert_state_agents(env.id, AgentStatus.paused, AgentStatus.down, AgentStatus.down)

    # one session
    ts1 = MockSession(uuid4(), env.id, {"agent1", "agent2"}, "ts1")
    await am._register_session(ts1, set(ts1.endpoint_names), datetime.datetime.now())
    await futures.proccess()
    assert len(am.sessions) == 1
    ts1.get_client().set_state.assert_called_with("agent2", enabled=True)
    ts1.get_client().reset_mock()
    await assert_state_agents(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.down, sid2=ts1.id)

    # alive
    await am._seen_session(ts1, set(ts1.endpoint_names))
    await futures.proccess()
    assert len(am.sessions) == 1
    await assert_state_agents(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.down, sid2=ts1.id)

    # second session
    ts2 = MockSession(uuid4(), env.id, {"agent3", "agent2"}, "ts2")
    await am._register_session(ts2, set(ts2.endpoint_names), datetime.datetime.now())
    await futures.proccess()
    assert len(am.sessions) == 2
    ts2.get_client().set_state.assert_called_with("agent3", enabled=True)
    ts2.get_client().reset_mock()
    await assert_state_agents(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.up, sid2=ts1.id, sid3=ts2.id)

    # expire first
    await am._expire_session(ts1, set(ts1.endpoint_names), datetime.datetime.now())
    await futures.proccess()
    assert len(am.sessions) == 1
    ts2.get_client().set_state.assert_called_with("agent2", enabled=True)
    ts2.get_client().reset_mock()
    await assert_state_agents(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.up, sid2=ts2.id, sid3=ts2.id)

    # expire second
    await am._expire_session(ts2, set(ts2.endpoint_names), datetime.datetime.now())
    await futures.proccess()
    assert len(am.sessions) == 0
    await assert_state_agents(env.id, AgentStatus.paused, AgentStatus.down, AgentStatus.down)


async def assert_agent_db_state(
    tid: UUID, nr_procs: int, nr_non_expired_procs: int, nr_agent_instances: int, nr_non_expired_instances: int
) -> typing.Callable:
    """
    The database log is updated asynchronously. This method waits until
    the desired database state is reached.
    """

    async def is_db_state_reached():
        result = await data.AgentProcess.get_list(environment=tid)
        if len(result) != nr_procs:
            return False
        result = await data.AgentProcess.get_list(environment=tid, expired=None)
        if len(result) != nr_non_expired_procs:
            return False
        result = await data.AgentInstance.get_list(tid=tid)
        if len(result) != nr_agent_instances:
            return False
        result = await data.AgentInstance.get_list(tid=tid, expired=None)
        if len(result) != nr_non_expired_instances:
            return False
        return True

    await retry_limited(is_db_state_reached, 10)


async def test_session_renewal(init_dataclasses_and_load_schema):
    """
    Agent got timeout but agent process was still running (e.g, network connectivity was disrupted).
    So the same agent process connects to the server with the same session id. This test verifies
    that the database state is updated correctly when an expired session is renewed.
    """
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="testenv", project=project.id)
    await env.insert()

    agent_manager = AgentManager()
    agent_manager._stopped = False
    agent_manager._stopping = False

    session_manager = SessionManager()
    sid = uuid4()
    tid = env.id
    endpoint = "vm1"
    session = Session(
        sessionstore=session_manager,
        sid=sid,
        hang_interval=1,
        timout=1,
        tid=tid,
        endpoint_names=[endpoint],
        nodename="test",
        disable_expire_check=True,
    )

    await assert_agent_db_state(tid, nr_procs=0, nr_non_expired_procs=0, nr_agent_instances=0, nr_non_expired_instances=0)
    await agent_manager._register_session(
        session=session, endpoint_names_snapshot=set(session.endpoint_names), now=datetime.datetime.now()
    )
    await assert_agent_db_state(tid, nr_procs=1, nr_non_expired_procs=1, nr_agent_instances=1, nr_non_expired_instances=1)
    await agent_manager._expire_session(
        session=session, endpoint_names_snapshot=set(session.endpoint_names), now=datetime.datetime.now()
    )
    await assert_agent_db_state(tid, nr_procs=1, nr_non_expired_procs=0, nr_agent_instances=1, nr_non_expired_instances=0)
    await agent_manager._register_session(
        session=session, endpoint_names_snapshot=set(session.endpoint_names), now=datetime.datetime.now()
    )
    await assert_agent_db_state(tid, nr_procs=1, nr_non_expired_procs=1, nr_agent_instances=1, nr_non_expired_instances=1)


async def test_fix_corrupted_database(init_dataclasses_and_load_schema):
    """
    When the database connection is lost, the agent session information might get
    corrupted. This inconsistency should be fixed when a new session is registered.
    This test case verifies this behavior.
    """
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="testenv", project=project.id)
    await env.insert()

    agent_manager = AgentManager()
    agent_manager._stopped = False
    agent_manager._stopping = False

    session_manager = SessionManager()
    sid = uuid4()
    tid = env.id
    endpoint = "vm1"
    session = Session(
        sessionstore=session_manager,
        sid=sid,
        hang_interval=1,
        timout=1,
        tid=tid,
        endpoint_names=[endpoint],
        nodename="node1",
        disable_expire_check=True,
    )

    await assert_agent_db_state(tid, nr_procs=0, nr_non_expired_procs=0, nr_agent_instances=0, nr_non_expired_instances=0)
    await agent_manager._register_session(
        session=session, endpoint_names_snapshot=set(session.endpoint_names), now=datetime.datetime.now()
    )
    await assert_agent_db_state(tid, nr_procs=1, nr_non_expired_procs=1, nr_agent_instances=1, nr_non_expired_instances=1)
    await agent_manager._expire_session(
        session=session, endpoint_names_snapshot=set(session.endpoint_names), now=datetime.datetime.now()
    )
    await assert_agent_db_state(tid, nr_procs=1, nr_non_expired_procs=0, nr_agent_instances=1, nr_non_expired_instances=0)
    await agent_manager._register_session(
        session=session, endpoint_names_snapshot=set(session.endpoint_names), now=datetime.datetime.now()
    )
    await assert_agent_db_state(tid, nr_procs=1, nr_non_expired_procs=1, nr_agent_instances=1, nr_non_expired_instances=1)
    await agent_manager._expire_session(
        session=session, endpoint_names_snapshot=set(session.endpoint_names), now=datetime.datetime.now()
    )
    await assert_agent_db_state(tid, nr_procs=1, nr_non_expired_procs=0, nr_agent_instances=1, nr_non_expired_instances=0)

    # Make database corrupt
    instances = await data.AgentInstance.get_list(tid=tid)
    assert len(instances) == 1
    await instances[0].update(expired=None)
    # Assert corruption
    await assert_agent_db_state(tid, nr_procs=1, nr_non_expired_procs=0, nr_agent_instances=1, nr_non_expired_instances=1)

    # Session registration should fix the inconsistency
    await agent_manager._register_session(
        session=session, endpoint_names_snapshot=session.endpoint_names, now=datetime.datetime.now()
    )
    await assert_agent_db_state(tid, nr_procs=1, nr_non_expired_procs=1, nr_agent_instances=1, nr_non_expired_instances=1)


@pytest.mark.parametrize("auto_start_agent", [False])  # prevent autostart to keep agent under control
async def test_session_creation_fails(server, environment, async_finalizer, caplog):
    """
    Verify that:
     * Session creation works correctly when the connectivity to the database works.
     * Session creation is refused when the connectivity to the database doesn't work.
       In that case the server state should stay consistent.
    """
    env_id = UUID(environment)
    agentmanager = server.get_slice(SLICE_AGENT_MANAGER)

    assert len(agentmanager.sessions) == 0

    a = NullAgent(environment=environment)
    await a.add_end_point_name("agent1")
    async_finalizer(a.stop)
    await a.start()

    # Wait until session is created
    await retry_limited(lambda: (env_id, "agent1") in agentmanager.tid_endpoint_to_session, 10)

    # Verify that the session is created correctly
    session = agentmanager.tid_endpoint_to_session[(env_id, "agent1")]
    session_manager = session._sessionstore
    assert len(agentmanager.sessions) == 1
    assert session in agentmanager.sessions.values()
    assert len(session_manager._sessions) == 1
    assert session in session_manager._sessions.values()
    assert "Heartbeat failed" not in caplog.text

    await a.stop()
    await session.expire(0)

    # Wait until session expired
    await retry_limited(lambda: len(agentmanager.tid_endpoint_to_session) == 0, 10)

    # Verify session expiry
    assert len(agentmanager.sessions) == 0
    assert len(agentmanager.tid_endpoint_to_session) == 0
    assert len(session_manager._sessions) == 0

    # Remove connectivity to the database
    await data.disconnect()
    caplog.clear()

    a = NullAgent(environment=environment)
    await a.add_end_point_name("agent1")
    await a.start()
    async_finalizer(a.stop)

    # Verify that session creation fails and server state is stays consistent
    await retry_limited(lambda: "Heartbeat failed" in caplog.text, 10)
    assert len(agentmanager.sessions) == 0
    assert len(agentmanager.tid_endpoint_to_session) == 0
    assert len(session_manager._sessions) == 0


async def test_agent_on_resume_actions(server, environment, client, agent) -> None:
    config.Config.set("config", "agent-deploy-interval", "0")
    config.Config.set("config", "agent-repair-interval", "0")
    agent_manager = server.get_slice(SLICE_AGENT_MANAGER)

    env_id = uuid.UUID(environment)
    env = await data.Environment.get_by_id(env_id)
    await agent_manager.ensure_agent_registered(env=env, nodename="agent1")
    await agent_manager.ensure_agent_registered(env=env, nodename="agent2")
    await agent_manager.ensure_agent_registered(env=env, nodename="agent3")
    # The keep_paused_on_resume and unpause_on_resume actions are only allowed when the environment is halted
    result = await client.agent_action(environment, name="agent1", action=AgentAction.keep_paused_on_resume.value)
    assert result.code == 403
    result = await client.agent_action(environment, name="agent1", action=AgentAction.unpause_on_resume.value)
    assert result.code == 403

    result = await client.agent_action(environment, name="agent1", action="unknown_action")
    assert result.code == 400

    result = await client.agent_action(environment, name="agent2", action=AgentAction.pause.value)
    assert result.code == 200

    async def assert_agents_on_resume_state(agent_states: dict[str, Optional[bool]]) -> None:
        for agent_name, on_resume in agent_states.items():
            agent_from_db = await data.Agent.get_one(environment=env_id, name=agent_name)
            assert agent_from_db.unpause_on_resume is on_resume

    async def assert_agents_paused_state(agent_states: dict[str, bool]) -> None:
        for agent_name, paused in agent_states.items():
            agent_from_db = await data.Agent.get_one(environment=env_id, name=agent_name)
            assert agent_from_db.paused is paused

    await assert_agents_on_resume_state({"agent1": None, "agent2": None, "agent3": None})
    await assert_agents_paused_state({"agent1": False, "agent2": True, "agent3": False})

    # Halt the environment and check the on_resume state
    result = await client.halt_environment(environment)
    assert result.code == 200

    await assert_agents_on_resume_state({"agent1": True, "agent2": False, "agent3": True})
    await assert_agents_paused_state({"agent1": True, "agent2": True, "agent3": True})

    result = await client.agent_action(environment, name="agent1", action=AgentAction.keep_paused_on_resume.value)
    assert result.code == 200

    result = await client.agent_action(environment, name="agent2", action=AgentAction.unpause_on_resume.value)
    assert result.code == 200

    # The on_resume actions don't start the agents immediately, just manipulate the 'unpause_on_resume' flag
    await assert_agents_paused_state({"agent1": True, "agent2": True, "agent3": True})
    await assert_agents_on_resume_state({"agent1": False, "agent2": True, "agent3": True})

    result = await client.resume_environment(environment)
    assert result.code == 200

    await assert_agents_paused_state({"agent1": True, "agent2": False, "agent3": False})
    await assert_agents_on_resume_state({"agent1": None, "agent2": None, "agent3": None})

    # The keep_paused_on_resume and unpause_on_resume actions are only allowed when the environment is halted
    result = await client.all_agents_action(environment, action=AgentAction.keep_paused_on_resume.value)
    assert result.code == 403
    result = await client.all_agents_action(environment, action=AgentAction.unpause_on_resume.value)
    assert result.code == 403

    result = await client.halt_environment(environment)
    assert result.code == 200
    await assert_agents_on_resume_state({"agent1": False, "agent2": True, "agent3": True})

    result = await client.all_agents_action(environment, action=AgentAction.unpause_on_resume.value)
    assert result.code == 200
    await assert_agents_on_resume_state({"agent1": True, "agent2": True, "agent3": True})

    result = await client.all_agents_action(environment, action=AgentAction.keep_paused_on_resume.value)
    assert result.code == 200
    await assert_agents_on_resume_state({"agent1": False, "agent2": False, "agent3": False})

    result = await client.agent_action(environment, name="agent3", action=AgentAction.unpause_on_resume.value)
    assert result.code == 200
    await assert_agents_on_resume_state({"agent1": False, "agent2": False, "agent3": True})

    result = await client.resume_environment(environment)
    assert result.code == 200

    await assert_agents_paused_state({"agent1": True, "agent2": True, "agent3": False})
    await assert_agents_on_resume_state({"agent1": None, "agent2": None, "agent3": None})


async def test_auto_started_agent_log_in_debug_mode(server, environment):
    """
    Test the logging of an autostarted agent
    """
    logdir = Config.get("config", "log-dir")
    log_file_path = f"{logdir}/agent-{environment}.log"  # Path to the log file

    def log_contains_debug_line():
        try:
            with open(log_file_path) as f:
                log_content = f.read()
                return "DEBUG    inmanta.protocol.endpoints Start transport for client agent" in log_content
        except Exception:
            return False

    await retry_limited(log_contains_debug_line, 10)


@pytest.mark.parametrize("auto_start_agent", [True])
async def test_process_already_terminated(server, environment):
    """
    This test case tests whether the termination of autostarted agents (processes) happens correctly,
    when one of the processes has already terminated.
    """
    env_id = UUID(environment)
    env = await data.Environment.get_by_id(env_id)

    autostarted_agent_manager = server.get_slice(SLICE_AUTOSTARTED_AGENT_MANAGER)
    await autostarted_agent_manager._ensure_scheduler(env=env.id)
    assert len(autostarted_agent_manager._agent_procs) == 1

    # Terminate process
    autostarted_agent_manager._agent_procs[env_id].process.terminate()

    # This call shouldn't raise an exception
    await autostarted_agent_manager._terminate_agents()


@pytest.mark.parametrize("auto_start_agent", [False])  # prevent autostart to keep agent under control
async def test_exception_occurs_while_processing_session_action(server, environment, async_finalizer, monkeypatch, caplog):
    """
    This test verifies that the consumer of the _session_listener_actions queue keeps working
    even when the an exception is thrown while handling an action.
    """
    agentmanager = server.get_slice(SLICE_AGENT_MANAGER)

    # Replace the _process_action() method with one that throws an excption
    old_process_action_function = agentmanager._process_action

    async def new_process_action_function(self, action: SessionAction, session: Session, timestamp: datetime):
        raise Exception("Failure")

    monkeypatch.setattr(agentmanager, "_process_action", new_process_action_function)

    assert len(agentmanager.sessions) == 0

    # Start agent
    a = NullAgent(environment=environment)
    await a.add_end_point_name("agent1")
    await a.start()
    async_finalizer(a.stop)

    # Verify that an exception is thrown an no session is created
    await retry_limited(lambda: "An exception occurred while handling session action" in caplog.text, 10)
    assert len(agentmanager.sessions) == 0
    await a.stop()

    # Put original method in place.
    monkeypatch.setattr(agentmanager, "_process_action", old_process_action_function)

    # Start new agent
    a = NullAgent(environment=environment)
    await a.add_end_point_name("agent1")
    await a.start()
    async_finalizer(a.stop)

    # Verify that session is created successfully -> Consumer didn't crash
    await retry_limited(lambda: len(agentmanager.sessions) == 1, 10)


async def test_error_handling_agent_fork(server, environment, monkeypatch):
    """
    Verifies resolution of issue: inmanta/inmanta-core#2777
    """
    exception_message = "The start of the agent failed"

    async def _dummy_fork_inmanta(
        self, args: list[str], outfile: Optional[str], errfile: Optional[str], cwd: Optional[str] = None
    ) -> subprocess.Process:
        raise Exception(exception_message)

    # Make the _fork_inmanta method raise an Exception
    monkeypatch.setattr(AutostartedAgentManager, "_fork_inmanta", _dummy_fork_inmanta)

    autostarted_agent_manager = server.get_slice(SLICE_AUTOSTARTED_AGENT_MANAGER)
    with pytest.raises(Exception) as excinfo:
        await autostarted_agent_manager._ensure_scheduler(env=environment, restart=True)

    assert exception_message in str(excinfo.value)


@pytest.mark.parametrize("no_agent", [True])
async def test_are_agents_active(server, client, environment, async_finalizer) -> None:
    """
    Ensure that the `AgentManager.are_agents_active()` method returns True when an agent
    is in the up or the paused state.
    """
    agentmanager = server.get_slice(SLICE_AGENT_MANAGER)
    agent_name = "agent1"
    env_id = UUID(environment)
    env = await data.Environment.get_by_id(env_id)

    # The agent is not started yet ->  it should not be active
    assert not await agentmanager.are_agents_active(tid=env_id, endpoints=[agent_name])

    # Start agent
    await agentmanager.ensure_agent_registered(env, agent_name)
    a = NullAgent(environment=environment)
    await a.add_end_point_name("agent1")
    async_finalizer(a.stop)
    await a.start()

    # Verify agent is active
    await retry_limited(agentmanager.are_agents_active, tid=env_id, endpoints=[agent_name], timeout=10)

    # Pause agent
    result = await client.agent_action(tid=env_id, name=agent_name, action=AgentAction.pause.value)
    assert result.code == 200, result.result

    # Ensure the agent is still active
    await retry_limited(agentmanager.are_agents_active, tid=env_id, endpoints=[agent_name], timeout=10)


async def test_heartbeat_different_session(server_pre_start, async_finalizer, caplog):
    """
    Verify that:
      - if the max_clients is reached, the heartbeat will still work as it is in a different pool
      - the max_clients option in the config changes the number of max_clients in a pool
      - debug logs concerning 'max_clients limit reached' logged by Tornado, are logged as inmanta warnings.
    """
    caplog.set_level(logging.WARNING)
    hanglock = asyncio.Event()

    async def unlock():
        hanglock.set()

    async_finalizer(unlock)

    @typedmethod(
        path="/test",
        operation="GET",
        client_types=["agent"],
        agent_server=True,
    )
    def test_method(number: int) -> ReturnValue[int]:  # NOQA
        """
        api endpoint that never returns
        """

    class TestSlice(ServerSlice):
        @handle(test_method)
        async def test_method_implementation(self, number: int) -> ReturnValue[int]:  # NOQA
            LOGGER.warning(f"HANG {number}")
            await hanglock.wait()
            return ReturnValue(response=number)

    Config.set("agent_rest_transport", "max_clients", "1")

    # This part is copied from the app.start_agent function. It needs to be called before the server starts.
    # We need to be able to create an environment before we can create an agent which we can't do using the start_agent function
    max_clients: int = Config.get("agent_rest_transport", "max_clients", "10")
    AsyncHTTPClient.configure(None, max_clients=max_clients)

    server = TestSlice(name="test_slice")

    ibl = InmantaBootloader(configure_logging=True)
    ctx = ibl.load_slices()

    for mypart in ctx.get_slices():
        ibl.restserver.add_slice(mypart)

    ibl.restserver.add_slice(server)

    await ibl.start()
    async_finalizer.add(server.stop)
    async_finalizer.add(ibl.stop)

    client = protocol.Client("client")

    result = await client.create_project("project-test")
    assert result.code == 200
    proj_id = result.result["project"]["id"]

    result = await client.create_environment(proj_id, "test", None, None)
    assert result.code == 200
    env_id = result.result["environment"]["id"]
    environment = await data.Environment.get_by_id(uuid.UUID(env_id))

    agent_manager = ibl.restserver.get_slice(SLICE_AGENT_MANAGER)

    a = NullAgent(environment=environment.id)
    await a.add_end_point_name("agent1")

    async_finalizer.add(a.stop)
    await a.start()

    # Wait until session is created
    await retry_limited(lambda: len(agent_manager.sessions) == 1, 10)

    # Have many connections in flight
    hangers = asyncio.gather(*(a._client.test_method(i) for i in range(5)))
    logging.warning("WAITING!")
    assert not hangers.done()

    def did_exceed_capacity():
        msg = "max_clients limit reached, request queued. 1 active, 2 queued requests."
        for record in caplog.records:
            if msg in record.message:
                if record.name == "inmanta.protocol.endpoints" and record.levelno == logging.WARNING:
                    return True
        return False

    await retry_limited(did_exceed_capacity, 10)
    caplog.set_level(logging.NOTSET)
    caplog.clear()
    LOGGER.info("Locked, Waiting for heartbeat")

    def still_sending_heartbeats():
        count = caplog.text.count("TRACE sending heartbeat for")
        return count > 2

    await retry_limited(still_sending_heartbeats, 10)
    LOGGER.info("Heartbeat good, unlocking")
    hanglock.set()
    await hangers


@pytest.mark.parametrize("halt_environment", (True, False))
async def test_pause_all_agents_doesnt_pause_environment(server, environment, client, halt_environment: bool) -> None:
    """
    Reproduces bug: https://github.com/inmanta/inmanta-core/issues/9081. Additionally verifies that halting the entire
    environment does halt the scheduler process.
    """
    env_id = UUID(environment)
    env = await data.Environment.get_by_id(env_id)
    agent_manager = server.get_slice(SLICE_AGENT_MANAGER)
    autostarted_agent_manager = server.get_slice(SLICE_AUTOSTARTED_AGENT_MANAGER)

    await agent_manager.ensure_agent_registered(env=env, nodename="agent1")

    async def wait_for_state(*, active: bool) -> None:
        async def session_state():
            return await agent_manager.are_agents_active(env_id, [const.AGENT_SCHEDULER_ID]) == active

        await retry_limited(lambda: len(autostarted_agent_manager._agent_procs) == (1 if active else 0), timeout=2)
        await retry_limited(session_state, timeout=2)

    await wait_for_state(active=True)

    agents = await data.Agent.get_list()
    assert len(agents) == 2
    agent_dct = {agent.name: agent for agent in agents}
    assert not agent_dct[const.AGENT_SCHEDULER_ID].paused
    assert not agent_dct["agent1"].paused

    await wait_for_state(active=True)
    # pause agents / halt environment
    result = (
        await client.halt_environment(environment)
        if halt_environment
        else await client.all_agents_action(environment, AgentAction.pause.value)
    )
    assert result.code == 200
    # scheduler should be down only if the environment was halted
    await wait_for_state(active=not halt_environment)

    agents = await data.Agent.get_list()
    assert len(agents) == 2
    agent_dct = {agent.name: agent for agent in agents}
    assert agent_dct[const.AGENT_SCHEDULER_ID].paused == halt_environment
    assert agent_dct["agent1"].paused

    # scheduler still down (if it was down before)
    await wait_for_state(active=not halt_environment)
    # unpause agents / resume environment
    result = (
        await client.resume_environment(environment)
        if halt_environment
        else await client.all_agents_action(environment, AgentAction.unpause.value)
    )
    assert result.code == 200
    # scheduler should be up again if it was halted before
    await wait_for_state(active=True)

    agents = await data.Agent.get_list()
    assert len(agents) == 2
    agent_dct = {agent.name: agent for agent in agents}
    assert not agent_dct[const.AGENT_SCHEDULER_ID].paused
    assert not agent_dct["agent1"].paused



"""
Copyright 2017 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import typing
import warnings
from re import error as RegexError

import ply.lex as lex

from inmanta.ast import LocatableString, Location, Range
from inmanta.ast.constraint.expression import Regex
from inmanta.ast.variables import Reference
from inmanta.parser import ParserException, ParserWarning

keyworldlist = [
    "typedef",
    "as",
    "entity",
    "extends",
    "end",
    "in",
    "implementation",
    "for",
    "matching",
    "index",
    "implement",
    "using",
    "when",
    "and",
    "or",
    "not",
    "true",
    "false",
    "import",
    "is",
    "defined",
    "dict",
    "null",
    "undef",
    "parents",
    "if",
    "else",
    "elif",
]
literals = [":", "[", "]", "(", ")", "=", ",", ".", "{", "}", "?", "*", "%"]
reserved = {k: k.upper() for k in keyworldlist}

# List of token names.   This is always required
tokens = [
    "INT",
    "FLOAT",
    "ID",
    "CID",
    "SEP",
    "STRING",
    "MLS",
    "CMP_OP",
    "REGEX",
    "REL",
    "PEQ",
    "RSTRING",
    "FSTRING",
    "PLUS_OP",
    "MINUS_OP",
    "DIVISION_OP",
    "DOUBLE_STAR",
] + sorted(list(reserved.values()))


def t_REGEX(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r"matching[\s]+/([^/\\\n]|\\.)+/"
    # We include the "matching" part in the regex, because a regex is demarcated by two slashes.
    # By including the "matching" part in the regex, we can make the distinction between a regex and
    # two subsequent division operations.
    index_first_slash_char = t.value.index("/")
    part_before_regex = t.value[:index_first_slash_char]
    regex_with_slashes_as_str = t.value[index_first_slash_char:]
    regex_as_str = regex_with_slashes_as_str[1:-1]
    value = Reference("self")  # anonymous value
    try:
        expr = Regex(value, regex_as_str)
        t.value = expr
        return t
    except RegexError as error:
        t.value = regex_with_slashes_as_str
        end = t.lexer.lexpos - t.lexer.linestart + 1
        (s, e) = t.lexer.lexmatch.span()
        start = end - (e - s) + index_first_slash_char

        line_nr_regex = t.lexer.lineno + part_before_regex.count("\n")
        r: Range = Range(t.lexer.inmfile, line_nr_regex, start, line_nr_regex, end)
        raise ParserException(r, t.value, f"Regex error in {t.value}: '{error}'")


def t_FSTRING(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r"f(\"([^\\\"\n]|\\.)*\")|f(\'([^\\\'\n]|\\.)*\')"
    t.value = safe_decode(token=t, warning_message="Invalid escape sequence in f-string.", start=2, end=-1)
    lexer = t.lexer

    end = lexer.lexpos - lexer.linestart + 1
    (s, e) = lexer.lexmatch.span()
    start = end - (e - s)

    t.value = LocatableString(
        t.value, Range(lexer.inmfile, lexer.lineno, start, lexer.lineno, end), lexer.lexpos, lexer.namespace
    )

    return t


def t_RSTRING(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r"r(\"([^\\\"\n]|\\.)*\")|r(\'([^\\\'\n]|\\.)*\')"
    t.value = t.value[2:-1]
    lexer = t.lexer

    end = lexer.lexpos - lexer.linestart + 1
    (s, e) = lexer.lexmatch.span()
    start = end - (e - s)

    t.value = LocatableString(
        t.value, Range(lexer.inmfile, lexer.lineno, start, lexer.lineno, end), lexer.lexpos, lexer.namespace
    )

    return t


def t_ID(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r"[a-zA-Z_][a-zA-Z_0-9-]*"
    t.type = reserved.get(t.value, "ID")  # Check for reserved words
    if t.value[0].isupper():
        t.type = "CID"
    lexer = t.lexer

    end = lexer.lexpos - lexer.linestart + 1
    (s, e) = lexer.lexmatch.span()
    start = end - (e - s)

    t.value = LocatableString(
        t.value, Range(lexer.inmfile, lexer.lineno, start, lexer.lineno, end), lexer.lexpos, lexer.namespace
    )
    return t


def t_SEP(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r"[:]{2}"
    return t


def t_REL(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r"--|->|<-"
    return t


def t_CMP_OP(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r"!=|==|>=|<=|<|>"
    return t


def t_PEQ(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r"[+]="
    return t


def t_COMMENT(t: lex.LexToken) -> None:  # noqa: N802
    r"\#.*?\n"
    t.lexer.lineno += 1
    t.lexer.linestart = t.lexer.lexpos


def t_MLS(t: lex.LexToken) -> lex.LexToken:
    r'"{3,5}([\s\S]*?)"{3,5}'

    value = safe_decode(token=t, warning_message="Invalid escape sequence in multi-line string.", start=3, end=-3)

    lexer = t.lexer
    match = lexer.lexmatch[0]
    lines = match.split("\n")
    start_line = lexer.lineno
    end_line = lexer.lineno + len(lines) - 1
    t.lexer.lineno = end_line
    (s, e) = lexer.lexmatch.span()
    start = lexer.lexpos - lexer.linestart - (e - s) + 1
    end = len(lines[-1]) + 1

    t.value = LocatableString(value, Range(lexer.inmfile, start_line, start, end_line, end), lexer.lexpos, lexer.namespace)

    return t


def t_FLOAT(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r"[-]?[0-9]*[.][0-9]+"
    t.value = float(t.value)
    return t


def t_INT(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r"[-]?[0-9]+"
    t.value = int(t.value)
    return t


def t_STRING(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r"(\"([^\\\"\n]|\\.)*\")|(\'([^\\\'\n]|\\.)*\')"

    t.value = safe_decode(token=t, warning_message="Invalid escape sequence in string.", start=1, end=-1)

    lexer = t.lexer

    end = lexer.lexpos - lexer.linestart + 1
    (s, e) = lexer.lexmatch.span()
    start = end - (e - s)

    t.value = LocatableString(
        t.value, Range(lexer.inmfile, lexer.lineno, start, lexer.lineno, end), lexer.lexpos, lexer.namespace
    )

    return t


def t_PLUS_OP(t: lex.LexToken) -> lex.LexToken:
    r"\+"
    return t


def t_MINUS_OP(t: lex.LexToken) -> lex.LexToken:
    r"-"
    return t


def t_DIVISION_OP(t: lex.LexToken) -> lex.LexToken:
    r"/"
    return t


def t_DOUBLE_STAR(t: lex.LexToken) -> lex.LexToken:
    r"\*\*"
    return t


# Define a rule so we can track line numbers
def t_newline(t: lex.LexToken) -> None:  # noqa: N802
    r"\n+"
    t.lexer.lineno += len(t.value)
    t.lexer.linestart = t.lexer.lexpos


# A string containing ignored characters (spaces and tabs)
t_ignore = " \t"
# Error handling rule


def t_ANY_error(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    lexer = t.lexer

    end = lexer.lexpos - lexer.linestart + 1
    char: str = t.value[0]
    r: Range = Range(lexer.inmfile, lexer.lineno, end, lexer.lineno, end + 1)
    raise ParserException(r, char, "Illegal character '%s'" % char)


# Build the lexer
lexer = lex.lex()


def safe_decode(token: lex.LexToken, warning_message: str, start: int = 1, end: int = -1) -> str:
    r"""
    Check for the presence of an invalid escape sequence (e.g. "\.") in the value attribute of a given token.
    This function assumes to be called from within a t_STRING or a t_MLS rule.

    - Python < 3.12 raises a DeprecationWarning when encountering an invalid escape sequence
    - Python 3.12 will raise a SyntaxWarning
    - Future versions will eventually raise a SyntaxError
    (see https://docs.python.org/3.12/whatsnew/3.12.html#other-language-changes )

    :param token: The token whose value we want to decode.
    :param warning_message: The warning message to display.
    :param start: Start of the value slice (To only decode the characters after the leading quotation mark(s))
    :param end: End of the value slice (To only decode the characters before the trailing quotation mark(s))

    :return: The token value as a python str.
    """

    try:
        # This first block will try to decode the value and turn any deprecation warning into an actual Exception.
        with warnings.catch_warnings():
            warnings.filterwarnings("error", message="invalid escape sequence", category=DeprecationWarning)
            value: str = bytes(typing.cast(str, token.value)[start:end], "utf_8").decode("unicode_escape")
    except DeprecationWarning:
        # If the first block did actually encounter an invalid escape sequence, we have to decode the value again, this time
        # ignoring this or any other python warning that has already been emitted, and raising a warning of our own.
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            value = bytes(typing.cast(str, token.value)[start:end], "utf_8").decode("unicode_escape")

        warnings.warn(
            ParserWarning(location=Location(file=token.lexer.inmfile, lnr=token.lexer.lineno), msg=warning_message, value=value)
        )

    return value



"""
Copyright 2025 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import importlib.metadata
from collections.abc import Callable, Sequence
from typing import Optional

import packaging.version
from mypy import nodes, types, typevars
from mypy.plugin import AttributeContext, Plugin

method_namespaces: Sequence[str] = [
    entry_point.value for entry_point in importlib.metadata.entry_points(group="inmanta.mypy.methods")
]


class ClientMethodsPlugin(Plugin):
    def get_additional_deps(self, file: nodes.MypyFile) -> list[tuple[int, str, int]]:
        """
        Make sure method namespaces are loaded when the Client module is loaded.
        """
        if file.fullname != "inmanta.protocol.endpoints":
            return []
        return [(10, namespace, -1) for namespace in method_namespaces]

    def get_attribute_hook(self, fullname: str) -> Optional[Callable[[AttributeContext], types.CallableType]]:
        """
        For dynamic method accesses on a client object, return a hook that resolves to the associated method type signature,
        with the return type wrapped in a ClientCall.
        """
        method: Optional[types.CallableType] = self._get_method(fullname)
        if method is None:
            return None

        def hook(ctx: AttributeContext) -> types.CallableType:
            # unwrap ReturnValue[T]
            default_return_type_flattened: types.Type = (
                method.ret_type.args[0]
                if (
                    isinstance(method.ret_type, types.Instance)
                    and method.ret_type.type.fullname == "inmanta.protocol.common.ReturnValue"
                )
                else method.ret_type
            )

            return_type: types.Type
            if fullname.startswith("inmanta.protocol.endpoints.TypedClient."):
                # TypedClient returns method's return type without wrapping it
                return_type = default_return_type_flattened
            elif fullname.startswith("inmanta.protocol.endpoints.SyncClient."):
                # SyncClient returns method's return type wrapped in a Result object
                result_type: Optional[types.Instance] = self._get_instance("inmanta.protocol.common.Result")
                assert result_type is not None
                return_type = result_type.copy_modified(args=[default_return_type_flattened])
            # normal clients return method's return type wrapped in ClientCall or PageableClientCall
            elif (
                isinstance(default_return_type_flattened, types.Instance)
                and default_return_type_flattened.type.fullname == "builtins.list"
            ):
                pageable_client_call: Optional[types.Instance] = self._get_instance(
                    "inmanta.protocol.common.PageableClientCall"
                )
                assert pageable_client_call is not None
                return_type = pageable_client_call.copy_modified(args=[default_return_type_flattened.args[0]])
            else:
                client_call: Optional[types.Instance] = self._get_instance("inmanta.protocol.common.ClientCall")
                assert client_call is not None
                return_type = client_call.copy_modified(args=[default_return_type_flattened])

            return method.copy_modified(ret_type=return_type)

        return hook

    def _get_method(self, fullname: str) -> Optional[types.CallableType]:
        """
        If the given fully qualified name is a method access on a client object, returns the type signature object for that
        method. Returns None otherwise.
        """
        client_attr: Optional[str] = next(
            (
                name
                for prefix in (
                    "inmanta.protocol.endpoints.Client.",
                    "inmanta.protocol.endpoints.SessionClient.",
                    "inmanta.protocol.endpoints.SyncClient.",
                    "inmanta.protocol.endpoints.TypedClient.",
                    "inmanta.server.protocol.LocalClient.",
                )
                if (name := fullname.removeprefix(prefix)) != fullname
            ),
            None,
        )

        if client_attr is None or "." in client_attr:
            return None

        node: Optional[nodes.SymbolTableNode] = next(
            (
                lookup
                for namespace in method_namespaces
                if (lookup := self.lookup_fully_qualified(f"{namespace}.{client_attr}")) is not None
            ),
            None,
        )

        if node is None:
            return None

        result: Optional[types.Type] = node.type
        if result is None or not isinstance(result, types.CallableType):
            return None

        return result

    def _get_instance(self, fullname: str) -> Optional[types.Instance]:
        """
        Returns a mypy.types.Instance for the given full name, if it exists.
        """
        node: Optional[nodes.SymbolTableNode] = self.lookup_fully_qualified(fullname)
        if node is None or not isinstance(node.node, nodes.TypeInfo):
            return None
        generic: types.Instance | types.TupleType = typevars.fill_typevars(node.node)
        if not isinstance(generic, types.Instance):
            return None
        return generic


def plugin(version: str) -> type[Plugin]:
    return (
        ClientMethodsPlugin
        if packaging.version.Version(version) >= packaging.version.Version("1.17")
        # fall back to default behavior for older versions with unknown compatibility
        else Plugin
    )



"""
Copyright 2025 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import contextlib
import json
import logging
import os
import re
import typing
from logging import DEBUG
from typing import Iterator, Optional
from uuid import UUID

import pytest

from inmanta import compiler, env, references, resources, util
from inmanta.agent.handler import PythonLogger
from inmanta.ast import (
    ExternalException,
    PluginTypeException,
    RuntimeException,
    TypingException,
    UnexpectedReference,
    WrappingRuntimeException,
)
from inmanta.data.model import ReleasedResourceDetails, ReleasedResourceState
from inmanta.export import ResourceDict
from inmanta.references import Reference, ReferenceCycleException, reference
from inmanta.util.dict_path import Mapping, MutableMapping
from utils import ClientHelper, log_contains

if typing.TYPE_CHECKING:
    from conftest import SnippetCompilationTest

# The purpose of this module is to test references (compiler, exporter and executor). This module requires the test module
# defined in tests/data/modules_v2/refs


@contextlib.contextmanager
def raises_wrapped(
    exc_tp: type[Exception],
    *,
    match: Optional[str] = None,
    outer_exception: type[WrappingRuntimeException | ExternalException] = WrappingRuntimeException,
) -> Iterator[None]:
    """
    Context manager wrapper around pytest.raises. Expects a WrappingRuntimeException to be raised, and asserts that it wraps
    the provided exception type and that its message matches the provided pattern.
    """
    with pytest.raises(outer_exception) as exc_info:
        yield
    assert isinstance(exc_info.value.__cause__, exc_tp)
    if match is not None:
        msg: str = str(exc_info.value.__cause__)
        assert re.search(match, msg) is not None, msg


def round_trip_resource(resource):
    serialized = resource.serialize()
    data = json.dumps(serialized, default=util.api_boundary_json_encoder)
    r = resources.Resource.deserialize(json.loads(data))

    # Test cloning includes caches
    clone = r.clone()
    assert clone._references_model == r._references_model
    assert clone._references == r._references
    assert clone._resolved == r._resolved
    r.resolve_all_references(PythonLogger(logging.getLogger("test.refs")))
    # old clone shares cache, but not resolved state, as it has not been mutated
    assert clone._references_model == r._references_model
    assert clone._references == r._references
    assert clone._resolved != r._resolved
    clone = r.clone()
    assert clone._references_model == r._references_model
    assert clone._references == r._references
    assert clone._resolved == r._resolved
    return r


def round_trip_ref(reference: Reference):
    serialized = reference.serialize()
    data = json.dumps(serialized, default=util.api_boundary_json_encoder)
    intermediate = references.ReferenceModel(**json.loads(data))
    references.reference.get_class(intermediate.type).deserialize(intermediate, None, logging.getLogger("test.refs"))


def test_references_in_model(
    snippetcompiler: "SnippetCompilationTest",
    modules_v2_dir: str,
    caplog,
) -> None:
    """Test the use of references in the model and if they produce the correct serialized form."""
    refs_module = os.path.join(modules_v2_dir, "refs")

    def assert_id(ref_list: list[references.ReferenceModel], ref_type: str, ref_id: str) -> references.ReferenceModel:
        """Assert that the reference of type `ref_type` has id `ref_id`"""
        refs = {str(ref.id): ref for ref in ref_list}
        assert ref_id in refs
        assert refs[ref_id].type == ref_type

        return refs[ref_id]

    snippetcompiler.setup_for_snippet(
        snippet="""
        import refs
        import refs::dc
        import std::testing

        subref = refs::create_string_reference(name="CWD")
        # Test equals method
        subref = refs::create_string_reference(name="CWD")

        test_ref = refs::dc::create_all_refs_dataclass_reference(maybe_ref_value=subref)
        # Test equals method, does not work on dataclasses
        # test_ref = refs::dc::create_all_refs_dataclass_reference(maybe_ref_value=refs::create_string_reference(name="CWD"))

        std::testing::NullResource(
            name="test",
            agentname="test",
            fail=refs::create_bool_reference(name=test_ref.maybe_ref_value),
        )
        """,
        install_v2_modules=[env.LocalPackagePath(path=refs_module)],
    )
    _, res_dict = snippetcompiler.do_export()
    assert len(res_dict) == 1
    serialized = res_dict.popitem()[1].serialize()

    # validate that our UUID is stable
    assert_id(serialized["references"], "refs::Bool", "0d3b10e5-3f6c-32ae-8662-02e5d46a59c7")
    assert_id(serialized["references"], "refs::dc::AllRefsDataclassReference", "1102e0ce-2f03-31a5-ac3e-ef1a6609daaf")
    assert_id(serialized["references"], "core::AttributeReference", "8eafdfaf-9734-3d72-85dd-423df232c28e")
    assert_id(serialized["references"], "refs::String", "a8ed8c4f-204a-3f7e-a630-e21cb20e9209")

    data = json.dumps(serialized, default=util.api_boundary_json_encoder)

    resource = resources.Resource.deserialize(json.loads(data))
    resource.resolve_all_references(PythonLogger(logging.getLogger("test.refs")))
    assert not resource.fail and resource.fail is not None

    with caplog.at_level(DEBUG):
        resource.get_reference_value(UUID("1102e0ce-2f03-31a5-ac3e-ef1a6609daaf"), PythonLogger(logging.getLogger("test.refs")))

    log_contains(caplog, "test.refs", DEBUG, "Using cached value for reference AllRefsDataclassReference('CWD')")


def test_undeclared_reference_in_map(
    snippetcompiler: "SnippetCompilationTest",
    modules_v2_dir: str,
    caplog,
) -> None:
    """
    Verify that a custom field map method transparently allows references.
    """
    refs_module = os.path.join(modules_v2_dir, "refs")

    snippetcompiler.setup_for_snippet(
        snippet="""
        import std::testing
        import refs

        refs::DeepResourceNoReferences(
           name="test",
           agentname="test",
           value=refs::create_string_reference(name="test"),
       )
        """,
        install_v2_modules=[env.LocalPackagePath(path=refs_module)],
    )
    _, resources = snippetcompiler.do_export()

    # Find the resource
    rbykey = {str(k): v for k, v in resources.items()}
    r = rbykey["refs::DeepResourceNoReferences[test,name=test]"]

    # ensure we have the nastiest path possible.
    assert r.mutators[0].args[2].value == "value.'inner.something'[0]"

    # make it resolve!
    r = round_trip_resource(r)

    # ensure it did resolve
    assert r.value["inner.something"][0] == "test"


async def test_deploy_end_to_end(
    snippetcompiler: "SnippetCompilationTest",
    modules_v2_dir: str,
    caplog,
    agent,
    client,
    clienthelper: ClientHelper,
    environment,
) -> None:
    refs_module = os.path.join(modules_v2_dir, "refs")
    snippetcompiler.setup_for_snippet(
        snippet="""
            import refs
            import refs::dc
            import std::testing

            test_ref = refs::create_bad_reference(name=refs::create_string_reference(name="CWD"))

            # bad ref will fail
            std::testing::NullResource(
                name="test",
                agentname="test",
                fail=refs::create_bool_reference(name=test_ref),
            )

            # good ref will work
            std::testing::NullResource(
                name="test2",
                agentname="test",
                fail=refs::create_bool_reference(name="testx"),
            )

            # Deeper mutator should also work
            refs::DeepResource(
                name="test3",
                agentname="test",
                value=refs::create_string_reference(name="testx"),
            )

            # and a dict of references
            refs::DictResource(
                name="test4",
                agentname="test",
                value={"Hello": refs::create_string_reference(name="World!")},
            )

            # and a dict ref
            refs::DeepResource(
                name="test4",
                agentname="test",
                value=refs::create_DictRef(value = refs::create_string_reference(name="TESTX")),
            )
            """,
        install_v2_modules=[env.LocalPackagePath(path=refs_module)],
    )

    await clienthelper.set_auto_deploy()
    version, resource = await snippetcompiler.do_export_and_deploy()

    # All resource must be dictpath compatible for the mutators to work!
    for resource in resource.values():
        assert isinstance(resource, Mapping)
        assert isinstance(resource, MutableMapping)

    await clienthelper.wait_for_released()
    await clienthelper.wait_for_deployed()
    result = await client.resource_details(environment, "std::testing::NullResource[test,name=test]")
    assert result.code == 200
    details = ReleasedResourceDetails(**result.result["data"])
    assert details.status == ReleasedResourceState.failed

    result = await client.resource_details(environment, "std::testing::NullResource[test,name=test2]")
    assert result.code == 200
    details = ReleasedResourceDetails(**result.result["data"])
    assert details.status == ReleasedResourceState.deployed

    result = await client.resource_details(environment, "refs::DeepResource[test,name=test3]")
    assert result.code == 200
    details = ReleasedResourceDetails(**result.result["data"])
    assert details.status == ReleasedResourceState.deployed
    result = await client.resource_logs(environment, "refs::DeepResource[test,name=test3]")
    assert result.code == 200
    assert [msg for msg in result.result["data"] if "Observed value: {'inner.something': 'testx'}" in msg["msg"]]

    result = await client.resource_details(environment, "refs::DictResource[test,name=test4]")
    assert result.code == 200
    details = ReleasedResourceDetails(**result.result["data"])
    assert details.status == ReleasedResourceState.deployed
    result = await client.resource_logs(environment, "refs::DictResource[test,name=test4]")
    assert result.code == 200
    assert [msg for msg in result.result["data"] if "Observed value: {'Hello': 'World!'}" in msg["msg"]]

    result = await client.resource_details(environment, "refs::DeepResource[test,name=test4]")
    assert result.code == 200
    details = ReleasedResourceDetails(**result.result["data"])
    assert details.status == ReleasedResourceState.deployed
    result = await client.resource_logs(environment, "refs::DeepResource[test,name=test4]")
    assert result.code == 200
    assert [msg for msg in result.result["data"] if "Observed value: {'inner.something': ['TESTX']}" in msg["msg"]]


def test_decoding_legacy_resources(snippetcompiler, modules_v2_dir):
    """The encoding for paths in mutable json changed after iso8.2.0 this test backward compat"""

    refs_module = os.path.join(modules_v2_dir, "refs")

    # set up project
    snippetcompiler.setup_for_snippet(
        "import refs",
        install_v2_modules=[env.LocalPackagePath(path=refs_module)],
        autostd=True,
    )

    compiler.do_compile()

    old_resource = r"""
        {
        "references": [
            {
                "type": "refs::DictMade",
                "args": [
                    {
                        "type": "mjson",
                        "name": "name",
                        "value": {
                            "name": null
                        },
                        "references": {
                            ".name": {
                                "type": "reference",
                                "name": ".name",
                                "id": "187fdec1-d0e8-3652-b672-e098be97a2d1"
                            }
                        }
                    }
                ],
                "id": "83a1981c-59d1-36d9-aa07-3351b5430732"
            },
            {
                "type": "refs::String",
                "args": [
                    {
                        "type": "literal",
                        "name": "name",
                        "value": "TESTX"
                    }
                ],
                "id": "187fdec1-d0e8-3652-b672-e098be97a2d1"
            }
        ],
        "purged": false,
        "value": {
            "inner": null
        },
        "name": "test4",
        "agentname": "test",
        "mutators": [
            {
                "type": "core::Replace",
                "args": [
                    {
                        "type": "resource",
                        "name": "resource",
                        "id": "refs::DeepResource[test,name=test4]"
                    },
                    {
                        "type": "reference",
                        "name": "value",
                        "id": "83a1981c-59d1-36d9-aa07-3351b5430732"
                    },
                    {
                        "type": "literal",
                        "name": "destination",
                        "value": "value.inner"
                    }
                ]
            }
        ],
        "send_event": false,
        "managed": true,
        "purge_on_delete": false,
        "receive_events": true,
        "requires": [],
        "version": 1,
        "id": "refs::DeepResource[test,name=test4],v=1"
    }
        """

    r = resources.Resource.deserialize(json.loads(old_resource))
    # Test cloning includes caches
    clone = r.clone()
    assert clone._references_model == r._references_model
    assert clone._references == r._references
    assert clone._resolved == r._resolved
    r.resolve_all_references(PythonLogger(logging.getLogger("test.refs")))
    # old clone shares cache, but not resolved state, as it has not been mutated
    assert clone._references_model == r._references_model
    assert clone._references == r._references
    assert clone._resolved != r._resolved
    clone = r.clone()
    assert clone._references_model == r._references_model
    assert clone._references == r._references
    assert clone._resolved == r._resolved


def test_references_in_plugins(snippetcompiler: "SnippetCompilationTest", modules_v2_dir: str) -> None:
    """
    Verify the validation of references in plugins, both on the boundary, and on access through a proxy.
    """
    refs_module = os.path.join(modules_v2_dir, "refs")

    # set up project
    snippetcompiler.setup_for_snippet(
        "import refs",
        install_v2_modules=[env.LocalPackagePath(path=refs_module)],
        autostd=True,
    )

    def run_snippet(snippet: str) -> None:
        """
        Wrapper around snippetcompiler.setup_for_snippet + runs compile and export.
        Passes appropriate options and prepends snippet with refs import.
        """
        snippetcompiler.setup_for_snippet(
            f"import refs import refs::dc import refs::plugins\n{snippet}", install_project=False, autostd=True
        )
        snippetcompiler.do_export()

    # Primitives

    # Scenario: plugin argument annotated as `object`
    run_snippet(snippet="refs::plugins::takes_obj('hello')")
    with pytest.raises(PluginTypeException, match="is a reference"):
        run_snippet(snippet="refs::plugins::takes_obj(refs::create_string_reference('name'))")
    ## accepts list with a reference in it and even allows access -> infeasible to check inside a black box
    ## -> this is not a strong requirement. The assertion is here simply to ensure the behavior remains stable
    run_snippet(snippet="refs::plugins::takes_obj(['hello', refs::create_string_reference('hello')])")
    run_snippet(snippet="refs::plugins::iterates_obj(['hello', refs::create_string_reference('hello')])")
    # Scenario: plugin argument annotated as `object | Reference`
    run_snippet(snippet="refs::plugins::takes_obj_ref('hello')")
    run_snippet(snippet="refs::plugins::takes_obj_ref(refs::create_string_reference('name'))")
    # Scenario: plugin argument annotated as `Reference`
    with pytest.raises(PluginTypeException, match=re.escape("Expected type: Reference[any]")):
        run_snippet(snippet="refs::plugins::takes_obj_ref_only('hello')")
    run_snippet(snippet="refs::plugins::takes_obj_ref_only(refs::create_string_reference('name'))")
    # Scenario: plugin argument annotated as `str`
    run_snippet(snippet="refs::plugins::takes_str('hello')")
    with pytest.raises(PluginTypeException, match="is a reference"):
        run_snippet(snippet="refs::plugins::takes_str(refs::create_string_reference('name'))")
    # Scenario: plugin argument annotated as `str | Reference[str]`
    run_snippet(snippet="refs::plugins::takes_str_ref('hello')")
    run_snippet(snippet="refs::plugins::takes_str_ref(refs::create_string_reference('name'))")
    with pytest.raises(PluginTypeException, match=re.escape("Expected type: Reference[string] | string")):
        # takes a str ref, not a bool ref
        run_snippet(snippet="refs::plugins::takes_str_ref(refs::create_bool_reference('name'))")
    # Scenario: plugin argument annotated with complex union including Reference[str | None]
    run_snippet(snippet="refs::plugins::takes_complex_union_or_ref('Hello World!')")
    run_snippet(snippet="refs::plugins::takes_complex_union_or_ref(42)")
    run_snippet(snippet="refs::plugins::takes_complex_union_or_ref(null)")
    run_snippet(snippet="refs::plugins::takes_complex_union_or_ref(refs::create_string_reference('Hello'))")
    with pytest.raises(PluginTypeException, match=re.escape("Expected type: Union[int,string,Reference[string?]]?")):
        run_snippet(snippet="refs::plugins::takes_complex_union_or_ref([1])")
    with pytest.raises(PluginTypeException, match=re.escape("Expected type: Union[int,string,Reference[string?]]?")):
        # takes a string reference, not a bool reference
        run_snippet(snippet="refs::plugins::takes_complex_union_or_ref(refs::create_bool_reference('Hello'))")
    # Scenario: plugin argument annotated with Reference[bool] | Reference[str]
    run_snippet(snippet="refs::plugins::takes_union_of_refs(refs::create_string_reference('Hello'))")
    run_snippet(snippet="refs::plugins::takes_union_of_refs(refs::create_bool_reference('Hello'))")
    with pytest.raises(PluginTypeException, match=re.escape("Expected type: Union[Reference[bool],Reference[string]]")):
        run_snippet(snippet="refs::plugins::takes_union_of_refs(refs::create_int_reference('Hello'))")
    with pytest.raises(PluginTypeException, match=re.escape("Expected type: Union[Reference[bool],Reference[string]]")):
        run_snippet(snippet="refs::plugins::takes_union_of_refs('Hello World!')")
    with pytest.raises(PluginTypeException, match=re.escape("Expected type: Union[Reference[bool],Reference[string]]")):
        run_snippet(snippet="refs::plugins::takes_union_of_refs(true)")
    # Scenario: plugin argument annotated as `Sequence[object]`
    run_snippet(snippet="refs::plugins::iterates_obj_list(['h', 'e'])")
    run_snippet(snippet="refs::plugins::iterates_obj_list(['h', 1])")
    with pytest.raises(PluginTypeException, match="contains a reference"):
        run_snippet(snippet="refs::plugins::iterates_obj_list(['h', refs::create_string_reference('name')])")
    # Scenario: plugin argument annotated as `Sequence[str]`
    run_snippet(snippet="refs::plugins::iterates_str_list(['h', 'e'])")
    with pytest.raises(PluginTypeException, match=re.escape("Expected type: string[]")):
        run_snippet(snippet="refs::plugins::iterates_str_list(['h', 1])")
    with pytest.raises(PluginTypeException, match="contains a reference"):
        run_snippet(snippet="refs::plugins::iterates_str_list(['h', refs::create_string_reference('name')])")
    # Scenario: plugin argument annotated as `Sequence[str | Reference[str]]`
    run_snippet(snippet="refs::plugins::iterates_str_ref_list(['h', 'e'])")
    with pytest.raises(PluginTypeException, match=re.escape("Expected type: (Reference[string] | string)[]")):
        run_snippet(snippet="refs::plugins::iterates_str_ref_list(['h', 1])")
    run_snippet(snippet="refs::plugins::iterates_str_ref_list(['h', refs::create_string_reference('name')])")
    # Scenario: plugin argument annotated as `Mapping[str, object]`
    run_snippet(snippet="refs::plugins::iterates_object_dict({'h': 'h', 'e': 'e'})")
    with pytest.raises(PluginTypeException, match="contains a reference"):
        run_snippet(snippet="refs::plugins::iterates_object_dict({'h': 'h', 'e': refs::create_string_reference('name')})")
    # Scenario: plugin argument annotated as `Mapping[str, object | Reference]`
    run_snippet(snippet="refs::plugins::iterates_object_ref_dict({'h': 'h', 'e': 'e'})")
    run_snippet(snippet="refs::plugins::iterates_object_ref_dict({'h': 'h', 'e': refs::create_string_reference('name')})")

    # Entities

    # Scenario: plugin annotated as `Entity`
    ## Entity annotation
    ### dataclasses allowed
    run_snippet(
        "refs::plugins::takes_entity(refs::dc::AllRefsDataclass(maybe_ref_value=refs::create_string_reference('hello')))"
    )
    run_snippet("refs::plugins::takes_entity(refs::dc::NoRefsDataclass())")
    ### references allowed, as long as no reference attribute is accessed
    run_snippet("refs::plugins::takes_entity(refs::dc::create_all_refs_dataclass_reference('hello'))")
    run_snippet("refs::plugins::takes_entity(refs::dc::create_no_refs_dataclass_reference())")

    # Scenario: plugin annotated as `Entity` accesses reference attribute during plugin execution
    ## no reference
    run_snippet("refs::plugins::read_entity_value(refs::dc::AllRefsDataclass(maybe_ref_value='Hello World!'))")
    run_snippet("refs::plugins::read_entity_list_value(refs::ListContainer(value=['Hello', 'World!']))")
    run_snippet(
        "refs::plugins::read_entity_list_head(refs::ListContainer(value=['Hello', refs::create_string_reference('hello')]))"
    )
    run_snippet("refs::plugins::read_entity_dict_value(refs::DictContainer(value={'Hello': 'World!', 'mykey': '42'}))")
    run_snippet(
        """
        refs::plugins::read_entity_dict_mykey(
            refs::DictContainer(value={'Hello': refs::create_string_reference('hello'), 'mykey': '42'})
        )
        """
    )
    ## reference
    ### in attribute
    with raises_wrapped(
        UnexpectedReference, match="Encountered unexpected reference .* Encountered at instance.maybe_ref_value"
    ):
        run_snippet(
            """\
            refs::plugins::read_entity_value(
                refs::dc::AllRefsDataclass(maybe_ref_value=refs::create_string_reference('hello'))
            )
            """
        )
    ### inside list attribute
    with raises_wrapped(UnexpectedReference, match=r"Encountered unexpected reference .* Encountered at instance\.value\[1\]"):
        run_snippet(
            """\
            refs::plugins::read_entity_list_value(refs::ListContainer(value=['Hello', refs::create_string_reference('hello')]))
            """
        )
    with raises_wrapped(UnexpectedReference, match=r"Encountered unexpected reference .* Encountered at instance\.value\[0\]"):
        run_snippet(
            "refs::plugins::read_entity_list_head(refs::ListContainer(value=[refs::create_string_reference('hello'), 'Hello']))"
        )
    ## inside list attribute but allowed
    run_snippet(
        """\
        refs::plugins::read_entity_list_value_or_ref(
            refs::ListContainer(value=['Hello', refs::create_string_reference('hello')])
        )
        """
    )
    ## inside list attribute: allowed on instance level but not on nested list level
    with raises_wrapped(UnexpectedReference, match=r"Encountered unexpected reference .* Encountered at instance\.value\[1\]"):
        run_snippet(
            """\
            refs::plugins::read_entity_list_value_allow_references_single_level(
                refs::ListContainer(value=['Hello', refs::create_string_reference('hello')])
            )
            """
        )
    ### inside dict attribute
    with raises_wrapped(
        UnexpectedReference, match=r"Encountered unexpected reference .* Encountered at instance\.value\['mykey'\]"
    ):
        run_snippet(
            """\
            refs::plugins::read_entity_dict_value(
                refs::DictContainer(value={'Hello': 'World!', 'mykey': refs::create_string_reference('hello')})
            )
            """
        )
    with raises_wrapped(
        UnexpectedReference, match=r"Encountered unexpected reference .* Encountered at instance\.value\['mykey'\]"
    ):
        run_snippet(
            """\
            refs::plugins::read_entity_dict_mykey(
                refs::DictContainer(value={'Hello': 'World!', 'mykey': refs::create_string_reference('hello')})
            )
            """
        )
    ## inside dict attribute but allowed
    run_snippet(
        """\
        refs::plugins::read_entity_dict_value_or_ref(
            refs::DictContainer(value={'Hello': 'World!', 'mykey': refs::create_string_reference('hello')})
        )
        """
    )
    ## reference, plugin explicitly allows it
    run_snippet(
        """\
        refs::plugins::read_entity_ref_value(
            refs::dc::AllRefsDataclass(maybe_ref_value=refs::create_string_reference('hello'))
        )
        """
    )
    ## reference access for list of entities
    ### no reference
    run_snippet(
        """\
        refs::plugins::read_list_entity_value(
            [
                refs::dc::AllRefsDataclass(maybe_ref_value='Hello'),
                refs::dc::AllRefsDataclass(maybe_ref_value='World!'),
            ]
        )
        """
    )
    ### reference
    with raises_wrapped(
        UnexpectedReference, match=r"Encountered unexpected reference .* Encountered at instances\[1\]\.maybe_ref_value"
    ):
        run_snippet(
            """\
            refs::plugins::read_list_entity_value(
                [
                    refs::dc::AllRefsDataclass(maybe_ref_value='Hello'),
                    refs::dc::AllRefsDataclass(maybe_ref_value=refs::create_string_reference('hello')),
                ]
            )
            """
        )
    ### reference, plugin explicitly allows it
    run_snippet(
        """\
        refs::plugins::read_list_entity_ref_value(
            [
                refs::dc::AllRefsDataclass(maybe_ref_value='Hello'),
                refs::dc::AllRefsDataclass(maybe_ref_value=refs::create_string_reference('hello')),
            ]
        )
        """
    )

    ## plugin argument annotated with Entity | DC => custom to_python
    #  -> verify that resulting dynamic proxy has appropriate path context
    with raises_wrapped(UnexpectedReference, match="Encountered unexpected reference .* Encountered at v.non_ref_value"):
        run_snippet(
            "refs::plugins::takes_union_with_dc(refs::NormalEntity(non_ref_value=refs::create_string_reference('test')))"
        )
    with raises_wrapped(UnexpectedReference, match=r"Encountered unexpected reference .* Encountered at v\[0\].non_ref_value"):
        run_snippet(
            "refs::plugins::takes_union_with_dc([refs::NormalEntity(non_ref_value=refs::create_string_reference('test'))])"
        )
    with raises_wrapped(
        UnexpectedReference, match=r"Encountered unexpected reference .* Encountered at v\['x'\].non_ref_value"
    ):
        run_snippet(
            "refs::plugins::takes_union_with_dc({'x': refs::NormalEntity(non_ref_value=refs::create_string_reference('test'))})"
        )

    # Scenario: plugin annotated as <dataclass> gets Reference[<dataclass>]
    ## dataclass type that does not support reference attrs
    ### plain dataclass
    run_snippet("refs::plugins::takes_no_refs_dataclass(refs::dc::NoRefsDataclass())")
    run_snippet(
        """\
        refs::plugins::takes_mixed_refs_dataclass(
            refs::dc::MixedRefsDataclass(maybe_ref_value=refs::create_string_reference('hello'))
        )
        """
    )
    ### basic inheritance
    run_snippet(
        """\
        refs::plugins::takes_no_refs_dataclass(
            refs::dc::MixedRefsDataclass(maybe_ref_value=refs::create_string_reference('hello'))
        )
        """
    )
    ### basic inheritance the wrong direction
    with pytest.raises(PluginTypeException, match=re.escape("Expected type: refs::dc::MixedRefsDataclass")):
        run_snippet("refs::plugins::takes_mixed_refs_dataclass(refs::dc::NoRefsDataclass())")
    ### references not allowed
    #### references to dataclass
    with pytest.raises(PluginTypeException, match="contains a reference"):
        run_snippet("refs::plugins::takes_no_refs_dataclass(refs::dc::create_no_refs_dataclass_reference())")
    with pytest.raises(PluginTypeException, match="contains a reference"):
        run_snippet("refs::plugins::takes_mixed_refs_dataclass(refs::dc::create_mixed_refs_dataclass_reference('hello'))")
    #### dataclass containing undeclared reference
    with pytest.raises(PluginTypeException, match="contains a reference"):
        run_snippet(
            """\
            refs::plugins::takes_no_refs_dataclass(
                refs::dc::NoRefsDataclass(non_ref_value=refs::create_string_reference('hello'))
            )
            """
        )
    ## dataclass type that supports reference attrs -> references are coerced
    ### references are coerced to dataclass of references
    run_snippet("refs::plugins::takes_all_refs_dataclass(refs::dc::create_all_refs_dataclass_reference('hello'))")
    ### references are coerced to dataclass of references, with inheritance
    run_snippet("refs::plugins::takes_dataclass(refs::dc::create_all_refs_dataclass_reference('hello'))")
    with pytest.raises(PluginTypeException, match="contains a reference"):
        run_snippet("refs::plugins::takes_dataclass(refs::dc::create_no_refs_dataclass_reference())")

    # Scenario: plugin annotated as Reference[<dataclass>]
    ## accepts a reference
    run_snippet("refs::plugins::takes_no_refs_dataclass_ref(refs::dc::create_no_refs_dataclass_reference())")
    ## rejects a dataclass
    with pytest.raises(PluginTypeException, match=re.escape("Expected type: Reference[refs::dc::NoRefsDataclass]")):
        run_snippet("refs::plugins::takes_no_refs_dataclass_ref(refs::dc::NoRefsDataclass())")

    # Scenario: plugin annotated as <dataclass> | Reference[<dataclass>]
    ## accepts either
    run_snippet("refs::plugins::takes_no_refs_dataclass_or_ref(refs::dc::create_no_refs_dataclass_reference())")
    run_snippet("refs::plugins::takes_no_refs_dataclass_or_ref(refs::dc::NoRefsDataclass())")
    run_snippet("refs::plugins::takes_mixed_refs_dataclass_or_ref(refs::dc::create_mixed_refs_dataclass_reference('hello'))")
    run_snippet("refs::plugins::takes_mixed_refs_dataclass_or_ref(refs::dc::MixedRefsDataclass(maybe_ref_value='hello'))")
    ## rejects other dataclasses / references to other dataclasses, except for inheritance
    ### allowed because MixedRefsDataclass is a child of NoRefsDataclass
    run_snippet("refs::plugins::takes_no_refs_dataclass_or_ref(refs::dc::MixedRefsDataclass(maybe_ref_value='hello'))")
    run_snippet("refs::plugins::takes_no_refs_dataclass_or_ref(refs::dc::create_mixed_refs_dataclass_reference('hello'))")
    ### not allowed because AllRefsDataclass is no child of NoRefsDataclass
    with pytest.raises(
        PluginTypeException,
        match=re.escape("Expected type: Reference[refs::dc::MixedRefsDataclass] | refs::dc::MixedRefsDataclass"),
    ):
        run_snippet("refs::plugins::takes_mixed_refs_dataclass_or_ref(refs::dc::AllRefsDataclass(maybe_ref_value='hello'))")
    with pytest.raises(
        PluginTypeException,
        match=re.escape("Expected type: Reference[refs::dc::MixedRefsDataclass] | refs::dc::MixedRefsDataclass"),
    ):
        run_snippet("refs::plugins::takes_mixed_refs_dataclass_or_ref(refs::dc::create_all_refs_dataclass_reference('hello'))")

    # Scenario: inheritance on return type
    ## declare generic, return specific
    ### takes_no_refs_dataclass only accepts specific no_refs_dataclass.
    ### We use it as an assertion that the other plugin returns the specific type in the DSL.
    run_snippet("refs::plugins::takes_no_refs_dataclass(refs::plugins::inheritance_return_specific())")
    ## declare generic reference, return specific reference
    run_snippet("refs::plugins::takes_no_refs_dataclass_ref(refs::plugins::inheritance_return_specific_ref())")

    # Scenario: plugin returns list attribute without reading elements, declares list[str] return
    ## allowed if attribute has no references
    run_snippet("refs::plugins::returns_entity_list(refs::ListContainer(value=['Hello', 'World!']))")
    ## error on return validation if attribute has references
    with pytest.raises(PluginTypeException, match=r"Return value .* has incompatible type\..*Expected type: string\[\]"):
        run_snippet(
            "refs::plugins::returns_entity_list(refs::ListContainer(value=['Hello', refs::create_string_reference('hello')]))"
        )
    ## allowed if plugin annotates reference in return type
    run_snippet(
        "refs::plugins::returns_entity_ref_list(refs::ListContainer(value=['Hello', refs::create_string_reference('hello')]))"
    )

    # Scenario: allow_reference_values() called on non-proxy list (e.g. list inside dataclass). Allowed, does nothing
    run_snippet("refs::plugins::allow_references_on_non_proxy()")

    # Scenario: accidental operators on references
    with raises_wrapped(NotImplementedError, outer_exception=ExternalException, match="is an inmanta reference, not a boolean"):
        run_snippet("refs::plugins::bool_on_reference(refs::create_string_reference('hello'))")
    # allowed by necessity because it makes error message very brittle otherwise
    run_snippet("refs::plugins::str_on_reference(refs::create_string_reference('hello'))")
    run_snippet("refs::plugins::repr_on_reference(refs::create_string_reference('hello'))")


def test_reference_cycle(snippetcompiler: "SnippetCompilationTest", modules_v2_dir: str) -> None:
    """Test the correct detection of reference cycles."""
    refs_module = os.path.join(modules_v2_dir, "refs")

    snippetcompiler.setup_for_snippet(
        snippet="""
        import refs
        import refs::dc
        import std::testing

        std::testing::NullResource(
            name="test",
            agentname="test",
            fail=refs::create_bool_reference_cycle(name="CWD"),
        )
        """,
        install_v2_modules=[env.LocalPackagePath(path=refs_module)],
    )

    with pytest.raises(
        ExternalException, match="Failed to get attribute 'fail' for export on 'std::testing::NullResource'"
    ) as e:
        snippetcompiler.do_export()
    assert isinstance(e.value.__cause__, ReferenceCycleException)
    assert "Reference cycle detected: StringReference() -> StringReference()" in str(e.value.__cause__)


def test_references_in_expression(snippetcompiler: "SnippetCompilationTest", modules_v2_dir: str) -> None:
    """Test that references are rejected in expressions"""
    refs_module = os.path.join(modules_v2_dir, "refs")

    snippetcompiler.setup_for_snippet(
        snippet="""
        import refs
        if refs::create_bool_reference_cycle(name="CWD"):
        end
        """,
        install_v2_modules=[env.LocalPackagePath(path=refs_module)],
        autostd=True,
    )

    with pytest.raises(
        RuntimeException,
        match=r"Invalid value `BoolReference\(StringReference\(\)\)`: "
        "the condition for an if statement can only be a boolean expression",
    ):
        snippetcompiler.do_export()


@pytest.mark.parametrize("agent", [True, False])
def test_references_in_resource_id(snippetcompiler: "SnippetCompilationTest", modules_v2_dir: str, agent: bool) -> None:
    """Test that references are rejected in the resource id value"""
    refs_module = os.path.join(modules_v2_dir, "refs")

    attr_assignments: str = "name='res1', agentname=ref" if agent else "name=ref, agentname='agent1'"
    snippetcompiler.setup_for_snippet(
        snippet=f"""
        import refs
        import refs::dc
        ref = refs::create_string_reference(name="CWD")

        refs::NullResource({attr_assignments})
        """,
        install_v2_modules=[env.LocalPackagePath(path=refs_module)],
        autostd=True,
    )

    with pytest.raises(
        resources.ResourceException,
        match=(
            f"Encountered reference in resource's {'agent' if agent else 'id'} attribute.*"
            f" Encountered at attribute '{'agent' if agent else ''}name'"
        ),
    ):
        snippetcompiler.do_export()


def test_references_in_wrong_resource(snippetcompiler: "SnippetCompilationTest", modules_v2_dir: str) -> None:
    """
    Test that we can't refer to other resources directly

    We only catch this on the remote side, as it is very hard to get into this situation.

    We use specially crafted resources, where the one to be constructed last creates a reference to the first one.
    """
    refs_module = os.path.join(modules_v2_dir, "refs")

    snippetcompiler.setup_for_snippet(
        snippet="""
        import refs
        value = refs::create_string_reference(name="CWD")

        refs::NullResource(name="test",agentname="test", fail=true)
        refs::NullResource(name="test2",agentname="test", fail=true)
        """,
        install_v2_modules=[env.LocalPackagePath(path=refs_module)],
        autostd=True,
    )

    _, res_dict = snippetcompiler.do_export()

    with pytest.raises(
        Exception,
        match=r"This resource refers to another resource refs::NullResource\[test,name=test2?\] instead of "
        r"itself refs::NullResource\[test,name=test2?\], this is not supported",
    ):
        for resource in res_dict.values():
            round_trip_resource(resource)


def test_references_in_index(snippetcompiler: "SnippetCompilationTest", modules_v2_dir: str) -> None:
    """Test that references are rejected in indexes"""
    refs_module = os.path.join(modules_v2_dir, "refs")

    snippetcompiler.setup_for_snippet(
        snippet="""
        import refs

        mystr = refs::create_string_reference("test")

        entity Test:
           string value
        end

        implement Test using std::none

        index Test(value)

        Test(value=mystr)
        """,
        install_v2_modules=[env.LocalPackagePath(path=refs_module)],
        autostd=True,
    )
    with pytest.raises(
        TypingException,
        match=r"Invalid value `StringReference\(\)` in index for attribute value: references can not be used in indexes",
    ):
        snippetcompiler.do_export()

    snippetcompiler.setup_for_snippet(
        snippet="""
          import refs

          mystr = refs::create_string_reference("test")

          entity Test:
             string value
          end

          implement Test using std::none

          index Test(value)

          a = Test[value=mystr]
          """,
        install_v2_modules=[env.LocalPackagePath(path=refs_module)],
        autostd=True,
    )
    with pytest.raises(
        TypingException,
        match=r"Invalid value `StringReference\(\)` in index for attribute value: references can not be used in indexes",
    ):
        snippetcompiler.do_export()


def test_ref_docs(snippetcompiler, monkeypatch, capsys):
    base_path = os.path.join(os.path.dirname(__file__), "..", "docs", "model_developers", "examples")

    def read_file(name: str) -> str:
        with open(os.path.join(base_path, name), "r") as fh:
            return fh.read()

    def run_for_example(name: str) -> ResourceDict:
        snippetcompiler.create_module(
            f"references{name}", read_file(f"references_{name}.cf"), read_file(f"references_{name}.py")
        )
        snippetcompiler.setup_for_snippet(f"import references{name}", autostd=True)
        _, resources = snippetcompiler.do_export()
        return resources

    resources = run_for_example(1)
    assert len(resources) == 1
    resource = next(iter(resources.values()))
    assert len(resource["mutators"]) != 0

    monkeypatch.setenv("test", "TEST VALUE")
    run_for_example(2)

    assert "TEST VALUE" in capsys.readouterr()[0]


def test_ref_serialization():

    @reference("test::Test")
    class TestReference(Reference[str]):

        def __init__(self, keys, none_attr=None):
            super().__init__()
            self.keys = keys
            self.none_attr = None

    round_trip_ref(TestReference({"a": "A"}))

    # Only clean json is allowed
    baddy = TestReference({"a": object()})
    with pytest.raises(ValueError):
        baddy.serialize()


def test_references_string_format(snippetcompiler: "SnippetCompilationTest", modules_v2_dir: str) -> None:
    """
    Verify the behavior of references in the string format operations.
    """
    refs_module = os.path.join(modules_v2_dir, "refs")

    # set up project
    snippetcompiler.setup_for_snippet(
        "import refs",
        install_v2_modules=[env.LocalPackagePath(path=refs_module)],
        autostd=True,
    )

    def run_snippet(snippet: str) -> None:
        """
        Wrapper around snippetcompiler.setup_for_snippet + runs compile and export.
        Passes appropriate options and prepends snippet with refs import.
        """
        snippetcompiler.setup_for_snippet(f"import refs\n{snippet}", install_project=False, autostd=True)
        snippetcompiler.do_export()

    # f-string format
    with pytest.raises(UnexpectedReference, match="Encountered reference in string format for variable `ref`"):
        run_snippet("ref = refs::create_string_reference('Hello') f'Hello {ref}'")
    # old-style string format
    with pytest.raises(UnexpectedReference, match="Encountered reference in string format for variable `{{ref}}`"):
        run_snippet("ref = refs::create_string_reference('Hello') 'Hello {{ref}}'")



"""
Copyright 2023 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import contextlib
import uuid
from typing import Optional

import pydantic
import pytest

from inmanta import compiler
from inmanta.validation_type import HashKeyContainer, _cachable_validate_type, regex_string, validate_type


@pytest.mark.parametrize(
    "attr_type,value,validation_parameters,is_valid",
    [
        ("pydantic.condecimal", 8, {"gt": 0, "lt": 10}, True),
        ("pydantic.condecimal", 8, {"gt": 0, "lt": 5}, False),
        ("pydantic.confloat", 1.5, {"multiple_of": 0.5}, True),
        ("pydantic.confloat", 1.5, {"multiple_of": 0.2}, False),
        ("pydantic.conint", 4, {"ge": 4}, True),
        ("pydantic.conint", 4, {"ge": 5}, False),
        ("pydantic.constr", "test123", {"regex": "^test.*$"}, True),
        ("pydantic.constr", "test123", {"regex": "^tst.*$"}, False),
        # constr with non-regex parameters
        ("pydantic.constr", "nomatch", {"regex": "^test.*$", "max_length": 10}, False),
        ("pydantic.constr", "testbuttoolong", {"regex": "^test.*$", "max_length": 10}, False),
        ("pydantic.constr", "testgood", {"regex": "^test.*$", "max_length": 10}, True),
        ("pydantic.constr", "noregex", {"max_length": 10}, True),
        ("pydantic.constr", "noregexbuttoolong", {"max_length": 10}, False),
        ("uuid.UUID", uuid.uuid4(), {}, True),
    ],
)
def test_type_validation(attr_type: str, value: str, validation_parameters: dict[str, object], is_valid: bool) -> None:
    """
    Test the behavior of the inmanta.validation_type.validate_type method.
    """
    validation_error: Optional[pydantic.ValidationError] = None
    try:
        validate_type(attr_type, value, validation_parameters)
    except (pydantic.ValidationError, ValueError) as e:
        validation_error = e
    assert (validation_error is None) is is_valid, validation_error


@pytest.mark.parametrize(
    "regex, value, is_valid",
    [
        ("^abc.*", "abc", True),
        ("^abc.*", "abcd", True),
        ("^abc.*", "xabc", False),
        ("^xabc.*", "xabc", True),
        ("^abc.*", "ab", False),
        (".*", True, False),
    ],
)
def test_regex_string(regex: str, value: object, is_valid: bool) -> None:
    with contextlib.nullcontext() if is_valid else pytest.raises(pydantic.ValidationError):
        pydantic.TypeAdapter(regex_string(regex)).validate_python(value)


def test_cache(snippetcompiler):
    # Verify we use the cache
    _cachable_validate_type.cache_clear()
    snippetcompiler.setup_for_snippet(
        """
        import std

        entity Tester:
          std::alfanum testfield
          std::any_http_url testfield3
        end

        implement Tester using std::none
        """
        + """
        Tester(testfield="A", testfield3="http://example.com/x")
        """
        * 20
    )

    compiler.do_compile()
    # For each field: One miss, 19 hits
    assert _cachable_validate_type.cache_info().hits == 19 * 2
    assert _cachable_validate_type.cache_info().misses == 2


@pytest.mark.parametrize(
    "attr_type,validation_parameters,schema",
    [
        ("pydantic.constr", {"regex": "^test.*$"}, {"pattern": "^test.*$", "type": "string"}),
        (
            "pydantic.constr",
            {"regex": "^test.*$", "max_length": 10},
            {"maxLength": 10, "pattern": "^test.*$", "type": "string"},
        ),
    ],
)
def test_schema(attr_type: str, validation_parameters: dict[str, object], schema: dict[str, str | int]) -> None:
    type_validator = _cachable_validate_type(attr_type, HashKeyContainer(validation_parameters))
    assert type_validator.json_schema() == schema



"""
Copyright 2017 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import dataclasses
import importlib
import inspect
import logging
import typing
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple, Union  # noqa: F401

import inmanta.ast.attribute
from inmanta import plugins
from inmanta.ast import (
    CompilerException,
    DataClassException,
    DataClassMismatchException,
    DuplicateException,
    Locatable,
    Location,
    MultiUnsetException,
    Named,
    Namespace,
    NotFoundException,
    RuntimeException,
    TypingException,
    WithComment,
)
from inmanta.ast.statements.generator import SubConstructor
from inmanta.ast.type import Any as inm_Any
from inmanta.ast.type import Float, NamedType, NullableType, Type
from inmanta.execute.runtime import Instance, QueueScheduler, Resolver, ResultVariable, dataflow
from inmanta.execute.util import AnyType, NoneValue
from inmanta.references import AttributeReference, PrimitiveTypes, Reference
from inmanta.types import DataclassProtocol

if TYPE_CHECKING:
    from inmanta.ast import Namespaced
    from inmanta.ast.attribute import Attribute, RelationAttribute  # noqa: F401
    from inmanta.ast.blocks import BasicBlock
    from inmanta.ast.statements import ExpressionStatement, Statement  # noqa: F401
    from inmanta.ast.statements.define import DefineAttribute, DefineImport, DefineIndex  # noqa: F401
    from inmanta.execute.runtime import ExecutionContext  # noqa: F401


LOGGER = logging.getLogger(__name__)


class Entity(NamedType, WithComment):
    """
    This class models a defined entity in the domain model of the configuration model.

    Each entity can contain attributes that are either data types or
    relations and each entity can inherit from parent entities.

    :param name: The name of this entity. This name can not be changed
        after this object has been created
    """

    def __init__(self, name: str, namespace: Namespace, comment: Optional[str] = None) -> None:
        NamedType.__init__(self)

        self.__name = name  # type: str

        self.__namespace = namespace

        self.parent_entities = []  # type: List[Entity]
        self.child_entities = []  # type: List[Entity]
        self._attributes = {}  # type: Dict[str,Attribute]

        self.implementations = []  # type: List[Implementation]
        self.implements = []  # type: List[Implement]
        self.implements_inherits = False

        # default values
        self.__default_values = {}  # type: Dict[str, DefineAttribute]

        self._index_def = []  # type: List[List[str]]
        self._indexes = []  # type: list[DefineIndex]
        self._index = {}  # type: Dict[str,Instance]
        self.index_queue = {}  # type: Dict[str,List[Tuple[ResultVariable, Statement]]]

        self._instance_list = set()  # type: Set[Instance]

        self.comment = comment

        self.normalized = False

        self._paired_dataclass: type[DataclassProtocol] | None = None
        self._paired_dataclass_field_types: dict[str, Type] = {}
        # Entities can be paired up to python dataclasses
        # If such a sibling exists, the type is kept here
        # Every instance of such entity can cary the associated python object in a slot called `DATACLASS_SELF_FIELD`

    def is_entity(self) -> bool:
        return True

    def normalize(self) -> None:
        self.normalized = True
        for attribute in self.__default_values.values():
            if attribute.default is not None:
                default_type: Type = attribute.type.get_type(self.namespace)
                try:
                    default_type.validate(attribute.default.as_constant())
                except RuntimeException as exception:
                    if exception.stmt is None or isinstance(exception.stmt, Type):
                        exception.set_statement(attribute)
                        exception.location = attribute.location
                    raise exception

        # check for duplicate relations in parent entities
        for name, my_attribute in self.get_attributes().items():
            if isinstance(my_attribute, inmanta.ast.attribute.RelationAttribute):
                for parent in self.parent_entities:
                    parent_attr = parent.get_attribute(name)
                    if parent_attr is not None:
                        raise DuplicateException(
                            my_attribute,
                            parent_attr,
                            f"Attribute name {name} is already defined in {parent_attr.entity.name},"
                            " unable to define relationship",
                        )

        # normalize implements but not implementations because they contain subblocks that require full type normalization first
        for i in self.implements:
            i.normalize()

        self.subc = [SubConstructor(self, i) for i in self.get_implements()]
        for sub in self.subc:
            sub.normalize()

        if self._paired_dataclass:
            self.pair_dataclass()

    def get_sub_constructor(self) -> list[SubConstructor]:
        return self.subc

    def get_implements(self) -> "List[Implement]":
        if self.implements_inherits:
            return self.implements + [i for p in self.parent_entities for i in p.get_implements()]
        else:
            return self.implements

    def add_default_value(self, name: str, value: "DefineAttribute") -> None:
        """
        Add a default value for an attribute
        """
        self.__default_values[name] = value

    def _get_own_defaults(self) -> "Dict[str, Optional[ExpressionStatement]]":
        return {k: v.default for k, v in self.__default_values.items() if v.default is not None or v.remove_default}

    def get_namespace(self) -> Namespace:
        """
        The namespace of this entity
        """
        return self.__namespace

    namespace = property(get_namespace)

    def __hash__(self) -> "int":
        """
        The hashcode of this entity is defined as the hash of the name
        of this entity
        """
        return hash(self.__name)

    def get_name(self) -> str:
        """
        Return the name of this entity. The name string has been
        internalised for faster dictionary lookups
        """
        return self.__name

    name = property(get_name)

    def get_full_name(self) -> str:
        """
        Get the full name of the entity
        """
        return self.__namespace.get_full_name() + "::" + self.__name

    def get_attributes(self) -> "Dict[str,Attribute]":
        """
        Get a set with all attributes that are defined in this entity
        """
        return self._attributes

    def set_attributes(self, attributes: "Dict[str,Attribute]") -> None:
        """
        Set a set of attributes that are defined in this entities
        """
        self._attributes = attributes

    attributes: "Dict[str,Attribute]" = property(get_attributes, set_attributes, None, None)

    def is_parent(self, parent_candidate: "Entity") -> bool:
        """
        Check if the given parent_candidate entity is a parent of this entity. Does not consider an entity its own parent.
        """
        if parent_candidate in self.parent_entities:
            return True
        else:
            for parent in self.parent_entities:
                if parent.is_parent(parent_candidate):
                    return True
        return False

    def get_all_parent_names(self) -> "List[str]":
        """
        Get a set with all parents of this entity
        """
        parents = [str(x) for x in self.parent_entities]
        for entity in self.parent_entities:
            parents.extend(entity.get_all_parent_names())

        return parents

    def get_all_parent_entities(self) -> "Set[Entity]":
        parents = [x for x in self.parent_entities]
        for entity in self.parent_entities:
            parents.extend(entity.get_all_parent_entities())
        return set(parents)

    def get_all_child_entities(self) -> "Set[Entity]":
        children = [x for x in self.child_entities]
        for entity in self.child_entities:
            children.extend(entity.get_all_child_entities())
        return set(children)

    def get_all_attribute_names(self) -> "List[str]":
        """
        Return a list of all attribute names, including parents
        """
        names = list(self._attributes.keys())

        for parent in self.parent_entities:
            names.extend(parent.get_all_attribute_names())

        return names

    def add_attribute(self, attribute: "Attribute") -> None:
        """
        Add an attribute to this entity. The attribute should not exist yet.
        """
        if attribute.name not in self._attributes:
            self._attributes[attribute.name] = attribute
        else:
            raise DuplicateException(
                self._attributes[attribute.name],
                attribute,
                "attribute '%s' already exists on entity '%s'" % (attribute.name, self.name),
            )

    def get_attribute(self, name: str) -> Optional["Attribute"]:
        """
        Get the attribute with the given name
        """
        if name in self._attributes:
            return self._attributes[name]
        else:
            for parent in self.parent_entities:
                attr = parent.get_attribute(name)
                if attr is not None:
                    return attr
        return None

    def has_attribute(self, attribute: str) -> bool:
        """
        Does the attribute already exist in this entity.
        """
        if attribute not in self._attributes:
            for parent in self.parent_entities:
                if parent.has_attribute(attribute):
                    return True

            return False
        else:
            return True

    def get_all_instances(self) -> "List[Instance]":
        """
        Return all instances of this entity
        """
        return list(self._instance_list)

    def add_instance(self, obj: "Instance") -> None:
        """
        Register a new instance
        """
        self._instance_list.add(obj)
        self.add_to_index(obj)

        for parent in self.parent_entities:
            parent.add_instance(obj)

    def get_instance(
        self,
        attributes: dict[str, object],
        resolver: Resolver,
        queue: QueueScheduler,
        location: Location,
        node: Optional[dataflow.InstanceNodeReference] = None,
    ) -> "Instance":
        """
        Return an instance of the class defined in this entity.
        If the corresponding node is not None, passes it on the instance.
        """
        out = Instance(self, resolver, queue, node)
        out.set_location(location)
        for k, v in attributes.items():
            out.set_attribute(k, v, location)

        self.add_instance(out)
        return out

    def is_subclass(self, superclass_candidate: "Entity", *, strict: bool = True) -> bool:
        """
        Check if self is a subclass of the given superclass_candidate.
        Does not consider entities a subclass of themselves in strict mode (the default).

        :param strict: Only return True for entities that are a strict subtype, i.e. not of the same type.
        """
        return (not strict and superclass_candidate == self) or self.is_parent(superclass_candidate)

    def issubtype(self, other: "Type") -> bool:
        return isinstance(other, inm_Any) or isinstance(other, Entity) and self.is_subclass(other, strict=False)

    def issupertype(self, other: "Type") -> bool:
        if not isinstance(other, Entity):
            raise NotImplementedError()
        return other.is_subclass(self, strict=True)

    def validate(self, value: object) -> bool:
        """
        Validate the given value
        """
        if isinstance(value, AnyType):
            return True

        if not isinstance(value, Instance):
            raise RuntimeException(None, f"Invalid type for value '{value}', should be type {self}")

        value_definition = value.type
        if not (value_definition is self or value_definition.is_subclass(self)):
            raise RuntimeException(None, f"Invalid class type for {value}, should be {self}")

        # Note on references:
        #
        # References to dataclasses are a special case in the sense that they are represented as plain instances in the DSL.
        # Their attributes get additional runtime validation on the boundary (see to_python()), so we can simply let them
        # pass validation here.
        #
        # Even non-reference instances may contain reference attributes so they get additional runtime validation
        # on plugin attribute access (see DynamicProxy.__getattr__). So these can be accepted here as well.

        return True

    def add_implementation(self, implement: "Implementation") -> None:
        """
        Register an implementation for this entity
        """
        self.implementations.append(implement)

    def add_implement(self, implement: "Implement") -> None:
        """
        Register an implementation for this entity
        """
        self.implements.append(implement)

    def __repr__(self) -> str:
        """
        The representation of this type
        """
        return "Entity(%s)" % (self.get_full_name())

    def __str__(self) -> str:
        """
        The pretty string of this type
        """
        return self.get_full_name()

    def __eq__(self, other: object) -> bool:
        """
        Override list eq method
        """
        if not isinstance(other, Entity):
            return False

        return self.name == other.name and self.namespace == other.namespace

    def add_index(self, attributes: list[str], index_def: "DefineIndex") -> None:
        """
        Add an index over the given attributes.
        """
        # duplicate check
        for index in self._index_def:
            if len(index) == len(attributes) and all((a == b for a, b in zip(index, attributes))):
                return

        self._index_def.append(sorted(attributes))
        self._indexes.append(index_def)
        for child in self.child_entities:
            child.add_index(attributes, index_def)

    def get_indices(self) -> list[list[str]]:
        return self._index_def

    def add_to_index(self, instance: Instance) -> None:
        """
        Update indexes based on the instance and the attribute that has
        been set
        """

        def index_value_gate(key: str, value: object) -> str:
            if isinstance(value, Reference):
                raise TypingException(
                    None,
                    f"Invalid value `{value}` in index for attribute {key} on instance {instance}: "
                    f"references can not be used in indexes.",
                )
            return repr(value)

        attributes = {k: v for k, v in instance.slots.items() if v.is_ready()}

        # check if an index entry can be added
        for index_attributes in self.get_indices():
            index_ok = True
            key = []
            for attribute in index_attributes:
                if attribute not in attributes:
                    index_ok = False
                else:
                    key.append(f"{attribute}={index_value_gate(attribute, attributes[attribute].get_value())}")

            if index_ok:
                keys = ", ".join(key)

                if keys in self._index and self._index[keys] is not instance:
                    raise DuplicateException(instance, self._index[keys], "Duplicate key in index. %s" % keys)

                self._index[keys] = instance

                if keys in self.index_queue:
                    for x, stmt in self.index_queue[keys]:
                        x.set_value(instance, stmt.location)
                    self.index_queue.pop(keys)

    def lookup_index(
        self, params: "List[Tuple[str,object]]", stmt: "Statement", target: "Optional[ResultVariable]" = None
    ) -> "Optional[Instance]":
        """
        Search an instance in the index.
        """
        all_attributes: list[str] = [x[0] for x in params]
        attributes: set[str] = set()
        for attr in all_attributes:
            if attr in attributes:
                raise RuntimeException(stmt, "Attribute %s provided twice in index lookup" % attr)
            attributes.add(attr)

        found_index = False
        for index_attributes in self.get_indices():
            if set(index_attributes) == attributes:
                found_index = True

        if not found_index:
            raise NotFoundException(
                stmt, self.get_full_name(), "No index defined on %s for this lookup: " % self.get_full_name() + str(params)
            )

        def coerce(key: str, t: Type, v: object) -> object:
            """
            Coerce float-typed values to float (e.g. 1 -> 1.0)
            """
            if isinstance(v, Reference):
                raise TypingException(
                    None, f"Invalid value `{v}` in index for attribute {key}: " f"references can not be used in indexes."
                )

            match t:
                case Float():
                    return t.cast(v)
                case NullableType(element_type=Float() as float):
                    return v if isinstance(v, NoneValue) else float.cast(v)
                case _:
                    return v

        key = ", ".join(
            [
                "%s=%s"
                % (
                    k,
                    repr(coerce(k, self.get_attribute(k).type, v)),
                )
                for k, v in sorted(params, key=lambda x: x[0])
            ]
        )
        if target is None:
            if key in self._index:
                return self._index[key]
            else:
                return None
        elif key in self._index:
            target.set_value(self._index[key], stmt.location)
        else:
            if key in self.index_queue:
                self.index_queue[key].append((target, stmt))
            else:
                self.index_queue[key] = [(target, stmt)]
        return None

    def get_default_values(self) -> "Dict[str,ExpressionStatement]":
        """
        Return the dictionary with default values
        """
        values = []  # type: List[Tuple[str,Optional[ExpressionStatement]]]

        # left most parent takes precedence
        for parent in reversed(self.parent_entities):
            values.extend(parent.get_default_values().items())

        # self takes precedence
        values.extend(self._get_own_defaults().items())
        # make dict, remove doubles
        dvalues = dict(values)
        # remove erased defaults
        return {k: v for k, v in dvalues.items() if v is not None}

    def get_default(self, name: str) -> "ExpressionStatement":
        """
        Get a default value for a given name
        """
        defaults = self.get_default_values()
        if name not in defaults:
            raise AttributeError(name)
        return defaults[name]

    def final(self, excns: list[CompilerException]) -> None:
        for key, indices in self.index_queue.items():
            for _, stmt in indices:
                excns.append(NotFoundException(stmt, key, f"No match in index on type {self.get_full_name()} with key {key}"))
        for _, attr in self.get_attributes().items():
            attr.final(excns)

    def get_double_defined_exception(self, other: "Namespaced") -> "DuplicateException":
        return DuplicateException(self, other, "Entity %s is already defined" % (self.get_full_name()))

    def get_location(self) -> Location:
        return self.location

    def pair_dataclass_stage1(self) -> None:
        """
        Attach the associated dataclass in the python domain

        should only be called on children of std::Dataclass

        Called early to make plugins able to resolve this type from python domain
        """
        # Find the dataclass name
        namespace = self.namespace.get_full_name()
        module_name = "inmanta_plugins." + namespace.replace("::", ".")
        # Find the dataclass
        dataclass_module = importlib.import_module(module_name)
        dataclass_raw = getattr(dataclass_module, self.name, None)
        self._paired_dataclass = dataclass_raw
        if dataclass_raw is not None:
            dataclass_raw._paired_inmanta_entity = self

    def pair_dataclass(self) -> None:
        """
        Validate the associated dataclass in the python domain

        should only be called on children of std::Dataclass
        should be called after normalization
        """
        assert self.normalized

        namespace = self.namespace.get_full_name()
        module_name = "inmanta_plugins." + namespace.replace("::", ".")
        dataclass_name = module_name + "." + self.name

        dataclass_raw = self._paired_dataclass
        if dataclass_raw is None:
            raise DataClassMismatchException(
                self,
                None,
                dataclass_name,
                f"The dataclass {self.get_full_name()} defined at {self.location} has no corresponding python dataclass. "
                "Dataclasses must have a python counterpart that is a frozen dataclass.",
            )

        if not dataclasses.is_dataclass(dataclass_raw):
            raise DataClassMismatchException(
                self,
                None,
                dataclass_name,
                f"The python object {module_name}.{dataclass_raw.__name__} associated to  {self.get_full_name()} "
                f"defined at {self.location} is not a dataclass. "
                "Dataclasses must have a python counterpart that is a frozen dataclass.",
            )
        dataclass: type[DataclassProtocol] = dataclass_raw
        if not dataclass.__dataclass_params__.frozen:
            raise DataClassMismatchException(
                self,
                None,
                dataclass_name,
                f"The python object {module_name}.{dataclass.__name__} associated to  {self.get_full_name()} "
                f"defined at {self.location} is not frozen. "
                "Dataclasses must have a python counterpart that is a frozen dataclass.",
            )

        # Validate fields, collect errors
        dc_fields = {f.name: f for f in dataclasses.fields(dataclass)}
        dc_types = typing.get_type_hints(dataclass)
        failures = []

        for rel_or_attr_name in self.get_all_attribute_names():
            rel_or_attr = self.get_attribute(rel_or_attr_name)
            match rel_or_attr:
                case inmanta.ast.attribute.RelationAttribute() as rel:
                    # No relations except for requires and provides
                    if not (rel.entity.get_full_name() == "std::Entity" and rel.name in ["requires", "provides"]):
                        failures.append(
                            f"a relation called {rel_or_attr_name} is defined at {rel.location}. "
                            "Dataclasses are not allowed to have relations"
                        )
                case inmanta.ast.attribute.Attribute() as attr:
                    if rel_or_attr_name not in dc_fields:
                        failures.append(
                            f"The attribute {rel_or_attr_name} has no counterpart in the python domain. "
                            "All attributes of a dataclasses must be identical in both the python and inmanta domain.",
                        )
                        continue
                    inm_type = attr.type_internal

                    # all own fields are primitive
                    if not inm_type.is_attribute_type():
                        failures.append(
                            f"The attribute {rel_or_attr_name} of type `{inm_type}` is not primitive. "
                            "All attributes of a dataclasses have to be of a primitive type.",
                        )
                        continue

                    dc_fields.pop(rel_or_attr_name)
                    # Type correspondence
                    try:
                        dsl_type = plugins.to_dsl_type(dc_types[rel_or_attr_name], self.location, self.namespace)
                        self._paired_dataclass_field_types[rel_or_attr_name] = dsl_type
                        if not inm_type.corresponds_to(dsl_type):
                            failures.append(
                                f"The attribute {rel_or_attr_name} does not have the same type as "
                                "the associated field in the python domain. "
                                "All attributes of a dataclasses must be identical in both the python and inmanta domain.",
                            )
                    except TypingException:
                        # Can't even convert the type
                        failures.append(
                            f"The attribute {rel_or_attr_name} does not have the same type as "
                            "the associated field in the python domain. "
                            "All attributes of a dataclasses must be identical in both the python and inmanta domain.",
                        )

        # Anything left in the dict has no counterpart
        for dcfield in dc_fields.keys():
            failures.append(
                f"The field {dcfield} doesn't exist in the inmanta domain. "
                "All attributes of a dataclasses must be identical in both the python and inmanta domain",
            )

        if failures:
            python_file = inspect.getfile(dataclass)
            _, python_lnr = inspect.getsourcelines(dataclass)
            msgs = "\n".join(" " * 4 + "-" + failure for failure in failures)
            raise DataClassMismatchException(
                self,
                dataclass,
                dataclass_name,
                f"The dataclass {self.get_full_name()} defined at {self.location} does not match"
                f" the corresponding python dataclass at {python_file}:{python_lnr}. {len(failures)} errors:\n" + msgs + "\n",
            )

        # Only std::none as implementation
        for implement in self.get_implements():
            for imp in implement.implementations:
                if imp.get_full_name() != "std::none":
                    raise DataClassException(
                        self,
                        f"The dataclass {self.get_full_name()} defined at {self.location} "
                        f"has an implementation other than 'std::none', defined at {implement.location}. "
                        "Dataclasses can only have std::none as implemenation.",
                    )
        # No index
        if self.get_indices():
            index_locations = ",".join([str(i.location) for i in self._indexes])
            raise DataClassException(
                self,
                f"The dataclass {self.get_full_name()} defined at {self.location} has an indexes "
                f"defined at {index_locations}. Dataclasses can not have any indexes.",
            )

    def get_paired_dataclass(self) -> Optional[type[object]]:
        return self._paired_dataclass

    def corresponds_to(self, type: Type) -> bool:
        if isinstance(type, inm_Any):
            return True
        return type is self

    def as_python_type_string(self) -> "str | None":
        if self._paired_dataclass is None:
            return None
        namespace = self.namespace.get_full_name()
        module_name = "inmanta_plugins." + namespace.replace("::", ".")
        dataclass_name: str = module_name + "." + self.name
        return dataclass_name

    def has_custom_to_python(self) -> bool:
        return self._paired_dataclass is not None

    def to_python(self, instance: object, *, path: str) -> object:
        if self._paired_dataclass is None:
            assert False, f"This class {self.get_full_name()} has no associated python type, this conversion is not supported"

        assert isinstance(instance, Instance)

        if instance.type is not self:
            # allow inheritance: delegate to child type
            return instance.type.to_python(instance, path=path)

        def domain_conversion(value: object) -> object:
            if isinstance(value, NoneValue):
                return None
            if isinstance(value, list):
                return [domain_conversion(v) for v in value]
            if isinstance(value, dict):
                return {k: domain_conversion(v) for k, v in value.items()}
            return value

        def create() -> object:
            # Convert values
            # All values are primitive, so this is trivial
            kwargs = {k: v.get_value() for k, v in instance.slots.items() if k not in ["self", "requires", "provides"]}
            for k, v in kwargs.items():
                assert k in self._paired_dataclass_field_types
                # dynamic validation, mostly in case of references, because they are allowed in the model while they have to be
                # declared (potentially nested) in the Python domain.
                self._paired_dataclass_field_types[k].validate(v)
            assert self._paired_dataclass is not None
            return self._paired_dataclass(**{k: domain_conversion(v) for k, v in kwargs.items()})

        if instance.dataclass_self is None:
            # Handle unsets
            unset = [v for k, v in instance.slots.items() if k not in ["self", "requires", "provides"] if not v.is_ready()]
            if unset:
                raise MultiUnsetException("Unset values when converting instance to dataclass", unset)

            instance.dataclass_self = create()

        if isinstance(instance.dataclass_self, Reference):
            LOGGER.debug(
                "Coercing dataclass reference `%s` to a dataclass with reference attributes to satisfy plugin type constraints",
                instance.dataclass_self,
            )
            # coerce if the dataclass definition is reference-compatible. Do not overwrite dataclass_self
            return create()
        else:
            # or simply return the linked dataclass instance
            return instance.dataclass_self

    def from_python(self, value: object, resolver: Resolver, queue: QueueScheduler, location: Location) -> Instance:
        """
        Construct the instance for the associated python object

        Should only be called on objects of the proper type,
         i.e. for which 'corresponds_to' returns True
         i.e. instances of the associated dataclass
        """
        assert self._paired_dataclass is not None  # make mypy happy

        def convert_none(x: object | None) -> object:
            return x if x is not None else NoneValue()

        def convert_to_attr_ref(name: str) -> AttributeReference[PrimitiveTypes]:
            ar: AttributeReference[PrimitiveTypes] = AttributeReference(
                reference=value,
                attribute_name=name,
            )
            attribute = self.get_attribute(name)
            assert attribute is not None  # Mypy
            ar._model_type = attribute.get_type()
            return ar

        if isinstance(value, Reference):
            instance = self.get_instance(
                {k.name: convert_to_attr_ref(k.name) for k in dataclasses.fields(self._paired_dataclass)},
                resolver,
                queue,
                location,
                None,
            )
        else:
            instance = self.get_instance(
                {k.name: convert_none(getattr(value, k.name)) for k in dataclasses.fields(value)},
                resolver,
                queue,
                location,
                None,
            )

        instance.dataclass_self = value
        # generate an implementation
        for stmt in self.get_sub_constructor():
            stmt.emit(instance, queue)
        return instance


class Implementation(NamedType):
    """
    A module functions as a grouping of objects. This can be used to create
    high level roles that do not have any arguments, or they can be used
    to create mixin like aspects.
    """

    def __init__(
        self, name: str, stmts: "BasicBlock", namespace: Namespace, target_type: str, comment: Optional[str] = None
    ) -> None:
        Named.__init__(self)
        self.name = name
        self.statements = stmts
        self.namespace = namespace
        self.target_type = target_type
        self.comment = comment

    def set_type(self, entity: Entity) -> None:
        self.entity = entity
        entity.add_implementation(self)

    def __repr__(self) -> str:
        return "Implementation(name = %s)" % self.name

    def normalize(self) -> None:
        try:
            self.statements.normalize()
        except RuntimeException as e:
            e.set_statement(self, False)
            raise
        except CompilerException as e:
            e.set_location(self.location)
            raise

    def get_full_name(self) -> str:
        return self.namespace.get_full_name() + "::" + self.name

    def get_namespace(self) -> Namespace:
        return self.namespace

    def get_double_defined_exception(self, other: "Namespaced") -> "DuplicateException":
        raise DuplicateException(
            self, other, f"Implementation {self.get_full_name()} for type {self.target_type} is already defined"
        )

    def get_location(self) -> Location:
        return self.location

    def as_python_type_string(self) -> "str | None":
        raise NotImplementedError("Implementations should not be arguments to plugins, this code is not expected to be called")

    def corresponds_to(self, type: Type) -> bool:
        raise NotImplementedError("Implementations should not be arguments to plugins, this code is not expected to be called")

    def to_python(self, instance: object, *, path: str) -> object:
        raise NotImplementedError("Implementations should not be arguments to plugins, this code is not expected to be called")


class Implement(Locatable):
    """
    Define an implementation of an entity in functions of implementations
    """

    comment: Optional[str]

    def __init__(self) -> None:
        Locatable.__init__(self)
        self.constraint = None  # type: ExpressionStatement
        self.implementations = []  # type: List[Implementation]
        self.comment = None  # type: str
        self.normalized = False

    def normalize(self) -> None:
        if self.normalized:
            return
        self.normalized = True
        self.constraint.normalize()
